# Copyright 1999-2017. Plesk International GmbH. All rights reserved.
"""
Mail restore worker for outgoing mails rate limiting feature.

Json input format definition:
{
    "subscriptions": [{
        "sysusers": [ "string /* sysuser name */" ],
        "out_limit": "int /* e.g. 32; -1 is unlimited, 0 is don't allow at all */",
        "allow_sendmail" : "bool /* true, false */",

        "domains": [{
            "name": "string /* e.g. domain.tld */",
            "is_main": "bool",
            "out_limit": "int",

            "vhost_id": "string /* e.g. e0b0b52c-cf6e-4c72-bc21-5573bb83fafe */",

            "aliases": [ "string /* domain aliases list */"],

            "mails": [{
                "name": "string /* email part before @ */",
                "out_limit": "int",
                "aliases": ["string /* email aliases list (also parts before @) */"]
            }]
        }]
    }],

    "enabled": "bool",
    "stats_poller_cron_job_period": "int default=60 /* cron job period in minutes from 1 to 60 */"
}
"""
import json, sys, subprocess, os.path
from optparse import OptionParser
import plesk_log
from plesk_outgoing_mail_db import OutgoingMailLimitsDb
from plesk_outgoing_mail_batch_helper import OutgoingMailBatchOperationsHelper
from plesk_outgoing_mail_controller import OutgoingMailController
from plesk_outgoing_mail_layout import INSTALL_SENDMAIL_WRAPPER_POSTFIX

log = plesk_log.getLogger(__name__)

def load_input_data():
    return json.load(sys.stdin)

def enable_outgoing(controller, enable, default_cron_job_period=None):
    if enable:
        controller.enable()
        if default_cron_job_period is not None:
            controller.install_cron_job(period=default_cron_job_period)
    else:
        controller.disable()

def parse_app_options():
    parser = OptionParser()
    parser.add_option("--fast", action="store_true")
    parser.add_option("--handlers-only", action="store_true")
    parser.add_option("--no-restart-service", action="store_true")
    parser.add_option("--start", action="store_true")
    parser.add_option("--continue", action="store_true", dest="cont")
    parser.add_option("--finish", action="store_true")

    options, args = parser.parse_args()

    if options.start or options.finish:
        options.cont = True
    if not any([options.start, options.cont, options.finish]):
        raise Exception("At least one of options --start, --continue, --finish should be specified")
    if len(args) > 0:
        raise Exception("Redundant arguments are specified: %s" % (args,))

    return options

def postfix_fix_sendmail_wrapper():
    if os.path.exists(INSTALL_SENDMAIL_WRAPPER_POSTFIX):
        subprocess.Popen([INSTALL_SENDMAIL_WRAPPER_POSTFIX], stdout=subprocess.PIPE, stderr=subprocess.PIPE).poll()

def main():
    try:
        opts = parse_app_options()
        data = load_input_data()

        if not (opts.fast or opts.handlers_only):
            db = OutgoingMailLimitsDb()
            if opts.start:
                db.recreate()

            batch_helper = OutgoingMailBatchOperationsHelper(db)
            for subscription in data["subscriptions"]:
                batch_helper.create_subscription_entries(subscription)
                batch_helper.set_subscription_limits(subscription)

        if opts.finish:
            control = OutgoingMailController()
            enable_outgoing(control, data["enabled"], data.get("stats_poller_cron_job_period", None))
            postfix_fix_sendmail_wrapper()

    except SystemExit:
        raise
    except Exception as ex:
        sys.stderr.write('%s\n' % ex)
        log.debug('This exception happened at:', exc_info=sys.exc_info())
        sys.exit(1)

if __name__ == "__main__":
    main()

# vim: ts=4 sts=4 sw=4 et :
