# Copyright 1999-2017. Plesk International GmbH. All rights reserved.
"""
Mail account restore utility.

Json input format definition:
{
    "mails": [ {
        "local_part":         "string assert='is_mailname_valid($value)' assertmsg='Mailname is not valid'",
        "domain_part":        "string assert='is_domain_name_valid($value)' assertmsg='Domain name is not valid'",
        "mail_aliases":       [ "string assert='is_mailname_valid($value)' assertmsg='Mail alias is not valid'" ],
        "dom_aliases":        [ "string assert='is_domain_name_valid($value)' assertmsg='Domain alias is not valid'" ],
        "account": {
            "type":           "string default='plain' assert='is_mail_password_type_valid($value)' assertmsg='Mail password type is not valid (should be either `plain` or `crypt`)'",
            "password":       "string"
        },
        "antivirus": {
            "virusfilter":    "string assert='is_av_valid($value)' assertmsg='Antivirus is incorrect (should be one of ``, `none`, `kav`, `drweb`)'",
            "direction":      "string default='' assert='is_av_direction_valid($value)' assertmsg='Antivirus direction is incorrect (should be one of ``, `incoming`, `outgoing`, `any`)'"
        },
        "mailbox": {
            "enabled":        "bool",
            "quota":          "int64 default=-1 assert='is_mbox_quota_valid($value)' assertmsg='Mailbox quota value is out of bounds'",
            "quota_unit":     "string default='B' assert='is_valid_size_unit($value)' assertmsg='Mailbox quota unit value is invalid'",
            "forwards":       [ "string assert='is_full_mailname_valid($value)' assertmsg='Forwarding mailname is not valid'" ] ,
            "mail_groups":    [ "string assert='is_full_mailname_valid($value)' assertmsg='Group mailname is not valid'" ]
        },
        "antispam": {
            "action":         "string assert='is_antispam_action_valid($value)' assertmsg='Antivirus action is invalid (should be one of ``, `move`, `text`, `delete`)'",
            "required_score": "int default='0'",
            "rewrite_header": "string default=''",
            "blacklist":      [ "string" ],
            "whitelist":      [ "string" ],
            "unblacklist":    [ "string" ],
            "unwhitelist":    [ "string" ]
        },
        "autoresponder": {
            "enabled":        "bool",
            "attachments":    [ "string" ],
            "keystr":         "string default=''",
            "key_where":      "string default=''",
            "subject":        "string default=''",
            "reply_to":       "string default=''",
            "forward_to":     "string default=''",
            "content_type":   "string default=''",
            "charset":        "string default=''",
            "text":           "string default=''",
            "ans_freq":       "int default='0'",
            "mem_limit":      "int default='0'"
        },
        "outgoing": {
            "limit":          "int"
        }
    } ]
}
"""
import json, sys, subprocess, os.path
from optparse import OptionParser
import plesk_log, plesk_config
from plesk_outgoing_mail_db import OutgoingMailLimitsDb

log = plesk_log.getLogger(__name__)

def load_input_data():
    return json.load(sys.stdin)

def restore_outgoing_limits(local_part, domain_part, out_limit):
    outgoing_db = OutgoingMailLimitsDb()
    if not outgoing_db.fetch_mailname(local_part, domain_part):
        outgoing_db.create_mailname(local_part, domain_part)
    outgoing_db.set_mailname_limits(local_part, domain_part, out_limit)

def run_cpp_restore_worker(json_str):
    product_root_d = "/usr/local/psa"
    try:
        product_root_d = plesk_config.get('PRODUCT_ROOT_D')
    except KeyError as ex:
        log.debug("No PRODUCT_ROOT_D in psa.conf: %s", ex)
    worker_bin = os.path.join(product_root_d, "admin", "sbin", "mailname_repair.worker")

    p = subprocess.Popen([worker_bin], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    p.communicate(input=json_str)
    if p.returncode != 0:
        raise Exception("Worker %s failed with exit status %d" % (worker_bin, p.returncode))

class MyParser(OptionParser):
    def format_epilog(self, formatter):
        return self.epilog

def main():
    try:
        parser = MyParser()
        parser.description = "Restore single mailbox."
        parser.epilog = """
Json input format definition:
{
    "mails": [ {
        "local_part":         "string assert='is_mailname_valid($value)' assertmsg='Mailname is not valid'",
        "domain_part":        "string assert='is_domain_name_valid($value)' assertmsg='Domain name is not valid'",
        "mail_aliases":       [ "string assert='is_mailname_valid($value)' assertmsg='Mail alias is not valid'" ],
        "dom_aliases":        [ "string assert='is_domain_name_valid($value)' assertmsg='Domain alias is not valid'" ],
        "account": {
            "type":           "string default='plain' assert='is_mail_password_type_valid($value)' assertmsg='Mail password type is not valid (should be either `plain` or `crypt`)'",
            "password":       "string"
        },
        "antivirus": {
            "virusfilter":    "string assert='is_av_valid($value)' assertmsg='Antivirus is incorrect (should be one of ``, `none`, `kav`, `drweb`)'",
            "direction":      "string default='' assert='is_av_direction_valid($value)' assertmsg='Antivirus direction is incorrect (should be one of ``, `incoming`, `outgoing`, `any`)'"
        },
        "mailbox": {
            "enabled":        "bool",
            "quota":          "int64 default=-1 assert='is_mbox_quota_valid($value)' assertmsg='Mailbox quota value is out of bounds'",
            "quota_unit":     "string default='B' assert='is_valid_size_unit($value)' assertmsg='Mailbox quota unit value is invalid'",
            "forwards":       [ "string assert='is_full_mailname_valid($value)' assertmsg='Forwarding mailname is not valid'" ] ,
            "mail_groups":    [ "string assert='is_full_mailname_valid($value)' assertmsg='Group mailname is not valid'" ]
        },
        "antispam": {
            "action":         "string assert='is_antispam_action_valid($value)' assertmsg='Antivirus action is invalid (should be one of ``, `move`, `text`, `delete`)'",
            "required_score": "int default='0'",
            "rewrite_header": "string default=''",
            "blacklist":      [ "string" ],
            "whitelist":      [ "string" ],
            "unblacklist":    [ "string" ],
            "unwhitelist":    [ "string" ]
        },
        "autoresponder": {
            "enabled":        "bool",
            "attachments":    [ "string" ],
            "keystr":         "string default=''",
            "key_where":      "string default=''",
            "subject":        "string default=''",
            "reply_to":       "string default=''",
            "forward_to":     "string default=''",
            "content_type":   "string default=''",
            "charset":        "string default=''",
            "text":           "string default=''",
            "ans_freq":       "int default='0'",
            "mem_limit":      "int default='0'"
        },
        "outgoing": {
            "limit":          "int"
        }
    } ]
}
"""
        parser.parse_args()
        data = load_input_data()
# run cpp restore worker first:
        log.debug("Run restore worker")
        run_cpp_restore_worker(json.dumps(data))

# restore outgoing limits:
        for mail in data['mails']:
            sys.stderr.write("Restoring outgoing limits for %s@%s\n" % (mail["local_part"], mail["domain_part"]))
            restore_outgoing_limits(mail["local_part"], mail["domain_part"], mail["outgoing"]["limit"])
    except SystemExit:
        raise
    except Exception as ex:
        sys.stderr.write('%s\n' % ex)
        log.error('This exception happened at:', exc_info=sys.exc_info())
        sys.exit(1)

if __name__ == "__main__":
    main()

# vim: ts=4 sts=4 sw=4 et :
