# Copyright 1999-2017. Plesk International GmbH. All rights reserved.
""" Class for enabling, disabling outgoing antispam feature """
from __future__ import print_function

import os
import subprocess
import json
import errno
import glob
import random
import ConfigParser

import plesk_log
import plesk_subprocess

import plesk_outgoing_mail_layout as layout


log = plesk_log.getLogger(__name__)


MAIL_HANDLERS_CONTROL = os.path.join(layout.ADMIN_SBIN_DIR, 'mail_handlers_control')
HTTPD_MODULES_CONTROL = os.path.join(layout.ADMIN_SBIN_DIR, 'httpd_modules_ctl')

def fetch_mail_handlers():
    """ Fetch list of dictionaries describing currently installed mail handlers.

        Example of single dictionary:
            {'address':    'all',
             'context':    'none',
             'enabled':    False,
             'executable': '/opt/psa/handlers/hooks/spf',
             'name':       'spf',
             'priority':   10,
             'protected':  False,
             'queue':      'before-queue',
             'type':       'global'}
    """
    args = [MAIL_HANDLERS_CONTROL, '--list', '--json']
    log.debug("Calling mail_handlers_control with: %s", args)
    json_data = plesk_subprocess.check_output(args)
    return json.loads(json_data)['handlers']

def _mail_handler_op(action, name, type_, queue,
                     mailname=None, priority=100, executable=None, context=None, enabled=False, persistent=False):
    """ Perform 'action' with specified mail handler by invoking mail_handlers_control utility.
        Raises subprocess.CalledProcessError if execution failed.
    """
    if action not in ('add', 'remove', 'enable', 'disable'):
        raise ValueError("action value '%s' is invalid" % action)

    args = {
        'name': name,
        'type': type_,
        'queue': queue,
    }

    if mailname:
        args['mailname'] = mailname

    if action == 'add':
        if not executable:
            raise ValueError("executable must be specified for '%s' action" % action)

        args.update(priority=priority, executable=executable)
        if context:
            args['context'] = context
        if enabled:
            args['enabled'] = None
        if not persistent:
            args['dont-preserve-on-restore'] = None

    args_list = [MAIL_HANDLERS_CONTROL, '--' + action]
    for opt, val in args.iteritems():
        args_list.append('--%s=%s' % (opt, val) if val else '--%s' % opt)

    log.debug("Calling mail_handlers_control with: %s", args_list)
    subprocess.check_call(args_list)

def add_mail_handler(name, type_, queue, executable, enabled=True, **kwargs):
    """ Add mail handler. See _mail_handler_op() for additional options and details. """
    _mail_handler_op('add', name, type_, queue, executable=executable, enabled=enabled, **kwargs)

def remove_mail_handler(name, type_, queue, mailname=None):
    """ Remove mail handler(s). See _mail_handler_op() for additional options and details. """
    _mail_handler_op('remove', name, type_, queue, mailname=mailname)

def _apache_modules_op(action, *modules):
    """ Enable or disable Apache modules.
        Raises subprocess.CalledProcessError if execution failed.
    """
    if action not in ('enable', 'disable'):
        raise ValueError("action value '%s' is invalid" % action)

    args = [HTTPD_MODULES_CONTROL, '--' + action, ','.join(modules)]
    log.debug("Calling httpd_modules_ctl with: %s", args)
    subprocess.check_call(args)

def install_system_cron_job(name, command, period):
    """ Install new or update existing system cron job. 'period' is in minutes. """
    if not name or '/' in name:
        raise ValueError("cron job name '%s' is invalid" % name)
    if not command:
        raise ValueError("cron job command is empty")
    if not isinstance(period, int) or not 0 < period <= 60:
        raise ValueError("cron job period '%s' is invalid: should be an integer from 1 to 60 minutes" % period)

    path = os.path.join(layout.SYSTEM_CRON_DIR, name)
    minutes = "%d-59/%d" % (random.randrange(period), period)
    with open(path, 'w') as fp:
        #     minute   hour dom  mon  dow  user    command
        print(minutes, "*", "*", "*", "*", "root", command, sep='\t', file=fp)

def remove_system_cron_job(name=None, name_glob=None):
    """ Remove system cron job(s) by its name or glob name. Does nothing if job or jobs matching glob don't exist. """
    def _remove_system_cron_file(path):
        try:
            os.unlink(path)
        except OSError as ex:
            log.debug("Failed to remove system cron job file '%s': %s", path, ex)
            if ex.errno != errno.ENOENT:
                raise

    if name is not None:
        if not name or '/' in name:
            raise ValueError("cron job name '%s' is invalid" % name)

        _remove_system_cron_file(os.path.join(layout.SYSTEM_CRON_DIR, name))

    elif name_glob is not None:
        if not name_glob or '/' in name_glob:
            raise ValueError("cron job name glob '%s' is invalid" % name_glob)
        
        path_glob = os.path.join(layout.SYSTEM_CRON_DIR, name_glob)
        for path in glob.glob(path_glob):
            _remove_system_cron_file(path)

    else:
        raise ValueError("neither cron job name, nor its glob is given")


class OutgoingMailSettings(ConfigParser.SafeConfigParser):
    INIT_SECTION = 'init'

    def __init__(self):
        # SafeConfigParser is an old style class, so no super() here
        ConfigParser.SafeConfigParser.__init__(self)
        self.read(layout.OUTGOING_MAIL_CFG)

    def get_features_list(self, default_value=[]):
        """ Default features list as a tuple. """
        try:
            return tuple(self.get(self.INIT_SECTION, 'features').split())
        except (ConfigParser.Error, ValueError) as ex:
            log.debug("Features list cannot be read from config, using default %s: %s" % (default_value, ex))
            return tuple(default_value)

    def get_extra_whitelist(self):
        """ Additional whitelisted user names. """
        try:
            return self.get(self.INIT_SECTION, 'whitelist').split()
        except (ConfigParser.Error, ValueError) as ex:
            log.debug("Extra whitelist cannot be read from config: %s" % ex)
            return []


class OutgoingMailController(object):
    HANDLER_CONTEXT_FEATURES = ('dont-disclose-sender', 'limit', 'debug', 'in-memory', 'defer-on-internal-errors',
                                'fix-sender-envelope', 'fix-sender-headers', 
                                'verify-sender-envelope', 'verify-sender-headers')
    DEFAULT_FEATURES = ('apache-vhost-id', 'cron', 'limit')
    
    DEFAULT_CRON_JOB_NAME = 'statistics-poller'
    DEFAULT_CRON_JOB_COMMAND = "%s %s %s >/dev/null 2>&1" % (layout.SW_ENGINE_BIN, 
                                                             layout.DMS_TASK_SCRIPT, layout.DMS_TASK_NAME)

    def __init__(self):
        self._installed_handlers = None
        self._settings = OutgoingMailSettings()

    def enable(self, features=None):
        """ Enable outgoing mail rate limiting. """
        if features:
            listed_manually, features = True,  features.split()
        else:
            listed_manually, features = False, self._settings.get_features_list(self.DEFAULT_FEATURES)

        context_features = self._filter_context_features(features)
        if context_features:
            self._enable_sendmail_wrapper()

        if 'apache-vhost-id' in features:
            self._enable_vhost_id_support()

        if 'cron' in features:
            self.install_cron_job()

        default_context_features = self._filter_context_features(self.DEFAULT_FEATURES)
        persistent = (listed_manually and context_features != default_context_features)
        context_str = ','.join(context_features)

        for queue in ('before-queue', 'before-sendmail'):
            if not self._has_handler_installed(layout.LIMIT_OUT_HOOK_NAME, 'global', queue) or persistent:
                remove_mail_handler(layout.LIMIT_OUT_HOOK_NAME, 'global', queue)
                add_mail_handler(layout.LIMIT_OUT_HOOK_NAME, 'global', queue, layout.LIMIT_OUT_HOOK_PATH,
                                 mailname='all-recipients', priority=5, context=context_str, persistent=persistent)
            else:
                log.debug("Previously customized %s %s %s handler was not modified. "
                          "To force, disable the feature first or manually specify features list.",
                          layout.LIMIT_OUT_HOOK_NAME, 'global', queue)

    def disable(self, features=None):
        """ Disable outgoing mail rate limiting. """
        if features:
            features = features.split()
        else:
            features = self.DEFAULT_FEATURES + self.HANDLER_CONTEXT_FEATURES + self._settings.get_features_list()

        if 'apache-vhost-id' in features:
            self._disable_vhost_id_support()

        if 'cron' in features:
            self._remove_all_cron_jobs()

        remove_mail_handler(layout.LIMIT_OUT_HOOK_NAME, 'global', 'before-queue')
        remove_mail_handler(layout.LIMIT_OUT_HOOK_NAME, 'global', 'before-sendmail')

    def install_cron_job(self, name=None, command=None, period=60):
        """ Install or modify cron job. """
        if name is None:
            name = self.DEFAULT_CRON_JOB_NAME
        if command is None:
            command = self.DEFAULT_CRON_JOB_COMMAND

        install_system_cron_job(layout.CRON_JOB_PREFIX + name, command, int(period))

    def remove_cron_job(self, name=None):
        """ Remove cron job with a given name. """
        if name is None:
            name = self.DEFAULT_CRON_JOB_NAME
        remove_system_cron_job(layout.CRON_JOB_PREFIX + name)

    def _remove_all_cron_jobs(self):
        """ Remove all cron jobs related to outgoing mail rate limiting. """
        remove_system_cron_job(name_glob=layout.CRON_JOB_PREFIX + "*")

    def _filter_context_features(self, features):
        """ Return set of features which are set via handlers' contexts. """
        return set([feature for feature in features if feature in self.HANDLER_CONTEXT_FEATURES])

    def _has_handler_installed(self, name, type_, queue, persistent=True):
        if self._installed_handlers is None:
            self._installed_handlers = fetch_mail_handlers()
        filtered = filter(lambda h:
                            h['name'] == name and
                            h['type'] == type_ and
                            h['queue'] == queue and
                            h['protected'] == persistent and
                            h['enabled'],
                          self._installed_handlers)
        return bool(filtered)

    def _enable_sendmail_wrapper(self):
        """ Enable sendmail wrapper. """
        try:
            os.unlink(layout.SENDMAIL_WRAPPER_DISABLED_FLAG_FILE)
        except OSError as ex:
            log.debug("Sendmail wrapper flag '%s' removal failed: %s", layout.SENDMAIL_WRAPPER_DISABLED_FLAG_FILE, ex)
            if ex.errno != errno.ENOENT:
                raise

    def _enable_vhost_id_support(self):
        """ Enable Apache mod_sysenv which is required to pass vhost-id values. """
        _apache_modules_op('enable', 'sysenv')

    def _disable_vhost_id_support(self):
        """ Disable Apache mod_sysenv which is required to pass vhost-id values. """
        _apache_modules_op('disable', 'sysenv')


if __name__ == '__main__':
    import doctest
    import sys

    sys.exit(doctest.testmod().failed)

# vim: ts=4 sts=4 sw=4 et :
