# Copyright 1999-2017. Plesk International GmbH. All rights reserved.
""" Classes for outgoing mails rate limiting SQLite DB support. """
import sys
import re
import pwd
import random

import plesk_log

from plesk_sqlite3_db import Sqlite3Db, sqlite3
from plesk_outgoing_mail_controller import OutgoingMailSettings
import plesk_outgoing_mail_layout as layout


log = plesk_log.getLogger(__name__)


def strip_re(string, pattern):
    """ Strip all occurrences of the pattern in string.
        >>> strip_re('''
        ... --a--b
        ... -----c''', r'^--')
        '\\na--b\\n---c'
    """
    return re.sub(r'(?m)' + pattern, "", string)

# IMPORTANT NOTES about indices and performance. Please read before modifying queries or indices.
# * Indices are used to speed up queries by increasing DB size and INSERT time.
# * SQLite creates implicit index for any PRIMARY KEY or UNIQUE fields.
# * Implicit indices are not created for foreign keys, usually you need to create an index on child key columns of
#   each foreign key constraint. See http://www.sqlite.org/foreignkeys.html#fk_indexes for details.
# * Adding or modifying an index, especially a complex one (with more than one column) may significantly increase
#   performance of certain queries, as well as significantly decrease them if not done properly. Therefore upon 
#   modification of indices or time-critical queries you MUST test performance.
# * Columns order in an index does matter. If an index lists columns (a,b,c,d), then only (parts of) queries that 
#   match fields in the exact order of any prefix of this sequence will benefit from the index. Moreover, all but 
#   but last match must use '=', 'IN' or 'IS NULL'. See http://www.sqlite.org/optoverview.html for more details
#   and examples.

class MailsDb(Sqlite3Db):
    MAILS_DB_SCHEMA = strip_re("""
        CREATE TABLE mails (
            id              INTEGER PRIMARY KEY,            -- rowid alias, see http://www.sqlite.org/autoinc.html
            name            VARCHAR NOT NULL,               -- mail name (part before '@')
            domain_id       INTEGER NOT NULL,               -- parent domain id
            out_limit       INTEGER NOT NULL DEFAULT -1,    -- out limit: -1 is unlimited, 0 not allowed, >0 limited

            UNIQUE (name, domain_id),
            FOREIGN KEY(domain_id) REFERENCES domains(id)
                ON UPDATE CASCADE
                ON DELETE CASCADE
        );

        CREATE INDEX idx_mails_fk ON mails(domain_id);

        CREATE TABLE domains (
            id              INTEGER PRIMARY KEY,
            name            VARCHAR NOT NULL UNIQUE,
            subscription_id INTEGER NOT NULL,               -- parent subscription id
            out_limit       INTEGER NOT NULL DEFAULT -1,

            FOREIGN KEY(subscription_id) REFERENCES subscriptions(id) 
                ON UPDATE CASCADE
                ON DELETE CASCADE
                DEFERRABLE INITIALLY DEFERRED               -- allow temporary FK inconsistency until transaction ends
        );

        CREATE INDEX idx_domains_fk ON domains(subscription_id);

        CREATE TABLE subscriptions (
            id              INTEGER PRIMARY KEY,
            main_domain_id  INTEGER NOT NULL,               -- id of main domain on the subscription
            out_limit       INTEGER NOT NULL DEFAULT -1,
            allow_sendmail  BOOLEAN DEFAULT 1,              -- 1 is true, 0 is false

            FOREIGN KEY(main_domain_id) REFERENCES domains(id) 
                ON UPDATE CASCADE
                ON DELETE CASCADE
                DEFERRABLE INITIALLY DEFERRED
        );

        CREATE INDEX idx_subscription_fk ON subscriptions(main_domain_id);

        CREATE TABLE sysusers (
            id              INTEGER PRIMARY KEY,            -- sysuser uid
            name            VARCHAR NOT NULL UNIQUE,        -- primary user alias name (mostly for logging)
            subscription_id INTEGER NOT NULL,               -- parent subscription id

            FOREIGN KEY(subscription_id) REFERENCES subscriptions(id)
                ON UPDATE CASCADE
                ON DELETE CASCADE
        );

        CREATE INDEX idx_sysusers_fk ON sysusers(subscription_id);

        CREATE TABLE whitelist (
            id              INTEGER PRIMARY KEY,            -- whitelisted system user uid
            name            VARCHAR NOT NULL UNIQUE         -- whitelisted system user name
        );
    """, r'^ {8}')

    def __init__(self, db_path=None):
        super(MailsDb, self).__init__(db_path)

    def _get_schema(self):
        schema = super(MailsDb, self)._get_schema()
        schema += self.MAILS_DB_SCHEMA
        return schema

    def _setup_connection(self, connection):
        """ Setup connection. Enable foreign keys support if possible. """
        connection.execute("PRAGMA foreign_keys = ON")

    def _post_create(self):
        super(MailsDb, self)._post_create()
        self._populate_whitelist_from(layout.DEFAULT_WHITELIST)

    def fetch_sysuser_by_uid(self, sysuser_uid):
        """ Fetch following information about a given sysuser and subscription it belongs to:
                sysuser_uid, sysuser_name, main_domain_name, out_limit, allow_sendmail.
            Returns dictionary with these fields or None.

            >>> mails_db.fetch_sysuser_by_uid(-1) is None
            True
        """
        return self.run("SELECT s.id AS _id, u.id AS sysuser_uid, u.name AS sysuser_name, "
                        "       d.name AS main_domain_name, s.out_limit, s.allow_sendmail "
                        "FROM sysusers u JOIN subscriptions s ON u.subscription_id = s.id "
                        "                JOIN domains d ON s.main_domain_id = d.id "
                        "WHERE u.id = ?", sysuser_uid).fetchone()

    def fetch_sysuser_by_name(self, sysuser_name):
        """ Fetch following information about a given sysuser and subscription it belongs to:
                sysuser_uid, sysuser_name, main_domain_name, out_limit, allow_sendmail.
            Returns dictionary with these fields or None.

            >>> mails_db.fetch_sysuser_by_name('non-existent-sysuser') is None
            True
        """
        return self.run("SELECT s.id AS _id, u.id AS sysuser_uid, u.name AS sysuser_name, "
                        "       d.name AS main_domain_name, s.out_limit, s.allow_sendmail "
                        "FROM sysusers u JOIN subscriptions s ON u.subscription_id = s.id "
                        "                JOIN domains d ON s.main_domain_id = d.id "
                        "WHERE u.name = ?", sysuser_name).fetchone()

    def fetch_sysuser(self, sysuser):
        """ Fetch following information about a given sysuser and subscription it belongs to:
                sysuser_uid, sysuser_name, main_domain_name, out_limit, allow_sendmail.
            'sysuser' argument can either be uid (integer) or user name (string).
            Returns dictionary with these fields or None.
        """
        if isinstance(sysuser, int):
            return self.fetch_sysuser_by_uid(sysuser)
        else:
            return self.fetch_sysuser_by_name(sysuser)

    def fetch_subscription(self, main_domain_name):
        """ Fetch following information about a given subscription:
              main_domain_name, out_limit, allow_sendmail.
            Returns dictionary with these fields or None.

            >>> mails_db.fetch_subscription('non-existent-subscription') is None
            True
        """
        return self.run("SELECT s.id AS _id, lower(d.name) AS main_domain_name, s.out_limit, s.allow_sendmail "
                        "FROM subscriptions s JOIN domains d ON s.main_domain_id = d.id "
                        "WHERE lower(d.name) = lower(?)",
                        main_domain_name).fetchone()

    def fetch_domain(self, domain_name):
        """ Fetch following information about a given domain:
              name, main_domain_name, out_limit.
            Returns dictionary with these fields or None.

            >>> mails_db.fetch_domain('non-existent-domain') is None
            True
        """
        return self.run("SELECT d.id AS _id, lower(d.name) AS name, lower(m.name) AS main_domain_name, d.out_limit "
                        "FROM domains d JOIN subscriptions s ON d.subscription_id = s.id "
                        "               JOIN domains m ON s.main_domain_id = m.id "
                        "WHERE lower(d.name) = lower(?)", domain_name).fetchone()

    def fetch_mailname(self, mail_name, domain_name):
        """ Fetch following information about a given mail address:
              mail_name, domain_name, out_limit.
            Returns dictionary with these fields or None.

            >>> mails_db.fetch_mailname('non-existent-mail', 'non-existent-domain') is None
            True
        """
        return self.run("SELECT m.id AS _id, lower(m.name) AS mail_name, (d.name) AS domain_name, m.out_limit "
                        "FROM mails m JOIN domains d ON m.domain_id = d.id "
                        "WHERE lower(m.name) = lower(?) AND lower(d.name) = lower(?)", mail_name, domain_name).fetchone()

    def list_mailnames_by_local_part(self, mail_name):
        """ Fetch following information about mail addresses with a given local part mail_name:
              mail_name, domain_name, out_limit.
            Returns list of dictionaries with these fields or [] if there are no matches.

            >>> mails_db.list_mailnames_by_local_part('non-existent-mail')
            []
        """
        return self.run("SELECT m.id AS _id, lower(m.name) AS mail_name, lower(d.name) AS domain_name, m.out_limit "
                        "FROM mails m JOIN domains d ON m.domain_id = d.id "
                        "WHERE lower(m.name) = lower(?)", mail_name).fetchall()

    def create_sysuser(self, sysuser_name, main_domain_name, sysuser_uid=None):
        """ Add sysuser to an existing subscription identified by its main domain name.
            Returns number of inserted rows.

            >>> mails_db.create_subscription('main-domain0.tld')
            2
            >>> mails_db.create_sysuser('sysuser0', 'main-domain0.tld', 100)
            1
            >>> skip_key(mails_db.fetch_sysuser('sysuser0'), '_id') == {
            ...         'sysuser_uid': 100,
            ...         'sysuser_name': 'sysuser0',
            ...         'main_domain_name': 'main-domain0.tld',
            ...         'out_limit': -1, 'allow_sendmail': 1}
            True
            >>> mails_db.fetch_sysuser('sysuser0') == mails_db.fetch_sysuser(100)
            True
            >>> mails_db.delete_sysuser('sysuser0')
            1
            >>> mails_db.create_sysuser('root', 'main-domain0.tld')
            1
            >>> mails_db.fetch_sysuser('root')['sysuser_uid'] == 0
            True
            >>> mails_db.create_sysuser('sysuser0_1', 'Main-Domain0.tld', 1001)
            1
            >>> skip_key(mails_db.fetch_sysuser('sysuser0_1'), '_id') == {
            ...         'sysuser_uid': 1001,
            ...         'sysuser_name': 'sysuser0_1',
            ...         'main_domain_name': 'main-domain0.tld',
            ...         'out_limit': -1, 'allow_sendmail': 1}
            True
        """
        if sysuser_uid is None:
            sysuser_uid = pwd.getpwnam(sysuser_name).pw_uid
        return self.run("INSERT INTO sysusers (id, name, subscription_id) "
                        "VALUES (?, ?, (SELECT s.id FROM subscriptions s JOIN domains d ON s.main_domain_id = d.id "
                                       "WHERE lower(d.name) = lower(?)))",
                        sysuser_uid, sysuser_name, main_domain_name).rowcount

    def create_subscription(self, main_domain_name):
        """ Create subscription and its main domain with default limits.
            Returns number of inserted rows.

            >>> mails_db.create_subscription('main-domain1.tld')
            2
            >>> skip_key(mails_db.fetch_subscription('main-domain1.tld'), '_id') == {
            ...         'main_domain_name': 'main-domain1.tld',
            ...         'out_limit': -1, 'allow_sendmail': 1}
            True
            >>> mails_db.create_subscription('Main-Domain-MixedCase1.tld')
            2
            >>> skip_key(mails_db.fetch_subscription('Main-Domain-MixedCase1.tld'), '_id') == {
            ...         'main_domain_name': 'main-domain-mixedcase1.tld',
            ...         'out_limit': -1, 'allow_sendmail': 1}
            True
        """
        with self.get_connection() as conn:
            log.debug("Executing queries 'INSERT INTO subscriptions (main_domain_id) ...' and "
                      "'INSERT INTO domains (name, subscription_id) ...' with main_domain_name='%s'",
                      main_domain_name)
            subscription_id = conn.execute("INSERT INTO subscriptions (main_domain_id) VALUES (-1)").lastrowid
            domain_id = conn.execute("INSERT INTO domains (name, subscription_id) VALUES (lower(?), ?)",
                                     (main_domain_name, subscription_id)).lastrowid
            conn.execute("UPDATE subscriptions SET main_domain_id = ? WHERE id = ?", (domain_id, subscription_id))
            return 2

    def create_domain(self, domain_name, main_domain_name):
        """ Create domain with default limits that belongs to a given subscription identified by its main domain.
            Returns number of inserted rows.

            >>> mails_db.create_subscription('main-domain2.tld')
            2
            >>> mails_db.create_domain('addon-domain2.tld', 'main-domain2.tld')
            1
            >>> skip_key(mails_db.fetch_domain('addon-domain2.tld'), '_id') == {
            ...         'name': 'addon-domain2.tld',
            ...         'main_domain_name': 'main-domain2.tld',
            ...         'out_limit': -1}
            True
            >>> mails_db.create_domain('Addon-Domain-MixedCase2.tld', 'Main-Domain2.tld')
            1
            >>> skip_key(mails_db.fetch_domain('Addon-Domain-MixedCase2.tld'), '_id') == {
            ...         'name': 'addon-domain-mixedcase2.tld',
            ...         'main_domain_name': 'main-domain2.tld',
            ...         'out_limit': -1}
            True
        """
        return self.run("INSERT INTO domains (name, subscription_id) "
                        "VALUES (lower(?), (SELECT s.id FROM subscriptions s JOIN domains d ON d.id = s.main_domain_id "
                                    "WHERE lower(d.name) = lower(?)))",
                        domain_name, main_domain_name).rowcount

    def create_mailname(self, mail_name, domain_name):
        """ Create mailname with default limits that belongs to a given domain.
            Returns number of inserted rows.

            >>> mails_db.create_subscription('main-domain3.tld')
            2
            >>> mails_db.create_mailname('mail', 'main-domain3.tld')
            1
            >>> skip_key(mails_db.fetch_mailname('mail', 'main-domain3.tld'), '_id') == {
            ...         'mail_name': 'mail', 
            ...         'domain_name': 'main-domain3.tld', 
            ...         'out_limit': -1}
            True
            >>> mails_db.create_mailname('MixedCaseMail', 'main-domain3.tld')
            1
            >>> skip_key(mails_db.fetch_mailname('MixedCaseMail', 'main-domain3.tld'), '_id') == {
            ...         'mail_name': 'mixedcasemail',
            ...         'domain_name': 'main-domain3.tld',
            ...         'out_limit': -1}
            True
        """
        return self.run("INSERT INTO mails (name, domain_id) VALUES (lower(?), (SELECT id FROM domains WHERE lower(name) = lower(?)))",
                        mail_name, domain_name).rowcount

    def delete_sysuser(self, sysuser_name):
        """ Delete sysuser by its name.
            Returns number of removed rows from "sysusers" table.

            >>> mails_db.create_subscription('main-domain13.tld')
            2
            >>> mails_db.create_sysuser('sysuser13', 'main-domain13.tld', 113)
            1
            >>> mails_db.delete_sysuser('sysuser13')
            1
            >>> mails_db.fetch_sysuser('sysuser13') is None
            True
            >>> mails_db.fetch_subscription('main-domain13.tld') is not None
            True
        """
        return self.run("DELETE FROM sysusers WHERE name = ?", sysuser_name).rowcount

    def delete_subscription(self, main_domain_name):
        """ Delete subscription and all its child objects.
            Returns number of removed rows from "subscriptions" table.

            >>> mails_db.create_subscription('main-domain4.tld')
            2
            >>> mails_db.create_sysuser('sysuser4', 'main-domain4.tld', 104)
            1
            >>> mails_db.create_mailname('mail', 'main-domain4.tld')
            1
            >>> mails_db.delete_subscription('main-domain4.tld')
            1
            >>> mails_db.fetch_mailname('mail', 'main-domain4.tld')
            >>> mails_db.fetch_domain('main-domain4.tld')
            >>> mails_db.fetch_sysuser('sysuser4')
            >>> mails_db.fetch_sysuser(104)
            >>> mails_db.fetch_subscription('main-domain4.tld')
            >>> mails_db.create_subscription('Main-Domain-MixedCase4.tld')
            2
            >>> mails_db.delete_subscription('Main-Domain-MixedCase4.tld')
            1
            >>> mails_db.fetch_domain('Main-Domain-MixedCase4.tld')
        """
        return self.run("DELETE FROM subscriptions WHERE main_domain_id = (SELECT id FROM domains WHERE lower(name) = lower(?))",
                        main_domain_name).rowcount

    def delete_domain(self, domain_name):
        """ Delete domain and all its child objects.
            Returns number of removed rows from "domains" table.

            >>> mails_db.create_subscription('main-domain5.tld')
            2
            >>> mails_db.create_sysuser('sysuser5', 'main-domain5.tld', 105)
            1
            >>> mails_db.create_mailname('mail', 'main-domain5.tld')
            1
            >>> mails_db.delete_domain('main-domain5.tld')
            1
            >>> mails_db.fetch_domain('main-domain5.tld')
            >>> mails_db.fetch_sysuser('sysuser5')
            >>> mails_db.fetch_sysuser(105)
            >>> mails_db.fetch_subscription('main-domain5.tld')
            >>> mails_db.create_subscription('Main-Domain-MixedCase5.tld')
            2
            >>> mails_db.create_sysuser('sysuser5_1', 'Main-Domain-MixedCase5.tld', 105)
            1
            >>> mails_db.create_mailname('mail', 'Main-Domain-MixedCase5.tld')
            1
            >>> mails_db.delete_domain('Main-Domain-MixedCase5.tld')
            1
            >>> mails_db.fetch_domain('Main-Domain-MixedCase5.tld')
            >>> mails_db.fetch_sysuser('sysuser5_1')
            >>> mails_db.fetch_sysuser(105)
            >>> mails_db.fetch_subscription('Main-Domain-MixedCase5.tld')
        """
        return self.run("DELETE FROM domains WHERE lower(name) = lower(?)", domain_name).rowcount

    def delete_mailname(self, mail_name, domain_name):
        """ Delete mailname.
            Returns number of removed rows from "mails" table.

            >>> mails_db.create_subscription('main-domain6.tld')
            2
            >>> mails_db.create_sysuser('sysuser6', 'main-domain6.tld', 106)
            1
            >>> mails_db.create_mailname('mail', 'main-domain6.tld')
            1
            >>> mails_db.delete_mailname('mail', 'main-domain6.tld')
            1
            >>> mails_db.fetch_mailname('mail', 'main-domain6.tld') is None
            True
            >>> mails_db.fetch_domain('main-domain6.tld') is not None
            True
            >>> mails_db.fetch_sysuser('sysuser6') is not None
            True
            >>> mails_db.fetch_sysuser(106) is not None
            True
            >>> mails_db.fetch_subscription('main-domain6.tld') is not None
            True
            >>> mails_db.create_mailname('MixedCaseMail', 'main-domain6.tld')
            1
            >>> mails_db.delete_mailname('MixedCaseMail', 'main-domain6.tld')
            1
            >>> mails_db.fetch_mailname('MixedCaseMail', 'main-domain6.tld') is None
            True
            >>> mails_db.create_subscription('Main-Domain-MixedCase6.tld')
            2
            >>> mails_db.create_sysuser('sysuser6_1', 'Main-Domain-MixedCase6.tld', 1061)
            1
            >>> mails_db.create_mailname('MixedCaseMail', 'Main-Domain-MixedCase6.tld')
            1
            >>> mails_db.delete_mailname('MixedCaseMail', 'Main-Domain-MixedCase6.tld')
            1
            >>> mails_db.fetch_mailname('MixedCaseMail', 'Main-Domain-MixedCase6.tld') is None
            True
        """
        return self.run("DELETE FROM mails WHERE lower(name) = lower(?) AND domain_id = (SELECT id FROM domains WHERE lower(name) = lower(?))",
                        mail_name, domain_name).rowcount

    def rename_sysuser(self, old_sysuser_name, new_sysuser_name):
        """ Rename sysuser.
            Returns number of affected rows.

            >>> mails_db.create_subscription('main-domain7.tld')
            2
            >>> mails_db.create_sysuser('sysuser7', 'main-domain7.tld', 107)
            1
            >>> mails_db.rename_sysuser('sysuser7', 'renamed-sysuser7')
            1
            >>> mails_db.fetch_sysuser_by_name('sysuser7') is None
            True
            >>> mails_db.fetch_sysuser_by_name('renamed-sysuser7') is not None
            True
            >>> mails_db.fetch_sysuser_by_uid(107) is not None
            True
        """
        return self.run("UPDATE sysusers SET name = ? WHERE name = ?", 
                        new_sysuser_name, old_sysuser_name).rowcount

    def rename_domain(self, old_domain_name, new_domain_name):
        """ Rename domain.
            Returns number of affected rows.

            >>> mails_db.create_subscription('main-domain8.tld')
            2
            >>> mails_db.rename_domain('main-domain8.tld', 'renamed-main-domain8.tld')
            1
            >>> mails_db.fetch_domain('main-domain8.tld') is None
            True
            >>> mails_db.fetch_domain('renamed-main-domain8.tld') is not None
            True
            >>> mails_db.rename_domain('renamed-main-domain8.tld', 'Renamed-MixedCase-Main-Domain8.tld')
            1
            >>> mails_db.fetch_domain('renamed-main-domain8.tld') is None
            True
            >>> mails_db.fetch_domain('Renamed-MixedCase-Main-Domain8.tld') is not None
            True
            >>> mails_db.rename_domain('Renamed-MixedCase-Main-Domain8.tld', 'Renamed2-MixedCase-Main-Domain8.tld')
            1
            >>> mails_db.fetch_domain('Renamed-MixedCase-Main-Domain8.tld') is None
            True
            >>> mails_db.fetch_domain('Renamed2-MixedCase-Main-Domain8.tld') is not None
            True
            >>> mails_db.rename_domain('Renamed2-MixedCase-Main-Domain8.tld', 'renamed2-main-domain8.tld')
            1
            >>> mails_db.fetch_domain('Renamed2-MixedCase-Main-Domain8.tld') is None
            True
            >>> mails_db.fetch_domain('renamed2-main-domain8.tld') is not None
            True
        """
        return self.run("UPDATE domains SET name = lower(?) WHERE lower(name) = lower(?)",
                        new_domain_name, old_domain_name).rowcount

    def rename_mailname(self, old_mail_name, new_mail_name, domain_name):
        """ Rename mailname.
            Returns number of affected rows.

            >>> mails_db.create_subscription('main-domain9.tld')
            2
            >>> mails_db.create_mailname('mail', 'main-domain9.tld')
            1
            >>> mails_db.rename_mailname('mail', 'renamed-mail', 'main-domain9.tld')
            1
            >>> mails_db.fetch_mailname('mail', 'main-domain9.tld') is None
            True
            >>> mails_db.fetch_mailname('renamed-mail', 'main-domain9.tld') is not None
            True
            >>> mails_db.rename_mailname('renamed-mail', 'Renamed-MixedCase-mail', 'main-domain9.tld')
            1
            >>> mails_db.fetch_mailname('renamed-mail', 'main-domain9.tld') is None
            True
            >>> mails_db.fetch_mailname('Renamed-MixedCase-mail', 'main-domain9.tld') is not None
            True
            >>> mails_db.rename_mailname('Renamed-MixedCase-mail', 'Renamed2-MixedCase-mail', 'Main-Domain9.tld')
            1
            >>> mails_db.fetch_mailname('Renamed-MixedCase-mail', 'main-domain9.tld') is None
            True
            >>> mails_db.fetch_mailname('Renamed2-MixedCase-mail', 'Main-Domain9.tld') is not None
            True
            >>> mails_db.rename_mailname('Renamed2-MixedCase-mail', 'renamed2-mail', 'main-domain9.tld')
            1
            >>> mails_db.fetch_mailname('Renamed2-MixedCase-mail', 'main-domain9.tld') is None
            True
            >>> mails_db.fetch_mailname('renamed2-mail', 'Main-Domain9.tld') is not None
            True
        """
        return self.run("UPDATE mails SET name = lower(?) "
                        "WHERE name = lower(?) AND domain_id = (SELECT id FROM domains WHERE lower(name) = lower(?))",
                        new_mail_name, old_mail_name, domain_name).rowcount

    @staticmethod
    def _normalize_limit(value, name):
        """ Returns ( named placeholder or field reference, integer value from [-1; +inf) ) pair 
            for given input limit value and name. 

            >>> MailsDb._normalize_limit(None, 'out_limit')
            ('out_limit', None)
            >>> MailsDb._normalize_limit(-100500, 'out_limit')
            (':out_limit', -1)
            >>> MailsDb._normalize_limit('123', 'out_limit')
            (':out_limit', 123)
            >>> MailsDb._normalize_limit('over 9000', 'out_limit')
            Traceback (most recent call last):
                ...
            ValueError: invalid literal for int() with base 10: 'over 9000'
        """
        if value is None:
            return name, value

        value = int(value)
        if value < 0:
            value = -1

        return ':' + name, value

    @staticmethod
    def _normalize_flag(value, name):
        """ Returns ( named placeholder or field reference, boolean value ) pair
            for given input flag value and name.

            >>> MailsDb._normalize_flag(None, 'allow_sendmail')
            ('allow_sendmail', None)
            >>> MailsDb._normalize_flag(True, 'allow_sendmail')
            (':allow_sendmail', True)
            >>> MailsDb._normalize_flag(False, 'allow_sendmail')
            (':allow_sendmail', False)
            >>> MailsDb._normalize_flag(1, 'allow_sendmail')
            Traceback (most recent call last):
                ...
            ValueError: Boolean value or None is expected as 'allow_sendmail' flag value
        """
        if value is None:
            return name, value

        if not isinstance(value, bool):
            raise ValueError("Boolean value or None is expected as '%s' flag value" % name)

        return ':' + name, value

    def set_subscription_limits(self, main_domain_name, out_limit=None, allow_sendmail=None):
        """ Set in and out limits for subscription. 
            Limit values will be cast to integer. If limit value is None, it will be left untouched.
            'allow_sendmail' value should be either boolean or None.
            Returns number of affected rows.

            >>> mails_db.create_subscription('main-domain10.tld')
            2
            >>> mails_db.set_subscription_limits('main-domain10.tld', out_limit='500', allow_sendmail=False)
            1
            >>> mails_db.fetch_subscription('main-domain10.tld')['out_limit']
            500
            >>> mails_db.fetch_subscription('main-domain10.tld')['allow_sendmail']
            0
            >>> mails_db.set_subscription_limits('main-domain10.tld', -1)
            1
            >>> mails_db.fetch_subscription('main-domain10.tld')['out_limit']
            -1
            >>> mails_db.fetch_subscription('main-domain10.tld')['allow_sendmail']
            0
            >>> mails_db.set_subscription_limits('Main-Domain10.tld', out_limit='700', allow_sendmail=True)
            1
            >>> mails_db.fetch_subscription('Main-Domain10.tld')['out_limit']
            700
            >>> mails_db.fetch_subscription('Main-Domain10.tld')['allow_sendmail']
            1
        """
        out_ph, out_limit = self._normalize_limit(out_limit, 'out_limit')
        asm_ph, allow_sendmail = self._normalize_flag(allow_sendmail, 'allow_sendmail')

        query = ("UPDATE subscriptions SET out_limit = %s, allow_sendmail = %s "
                 "WHERE id = (SELECT subscription_id FROM domains WHERE lower(name) = lower(:name))" % (out_ph, asm_ph))
        return self.run(query,
                        out_limit=out_limit, allow_sendmail=allow_sendmail, 
                        name=main_domain_name).rowcount

    def set_domain_limits(self, domain_name, out_limit=None):
        """ Set in and out limits for domain. 
            Limit values will be cast to integer. If limit value is None, it will be left untouched.
            Returns number of affected rows.

            >>> mails_db.create_subscription('main-domain11.tld')
            2
            >>> mails_db.set_domain_limits('main-domain11.tld', 123)
            1
            >>> mails_db.fetch_domain('main-domain11.tld')['out_limit']
            123
            >>> mails_db.set_domain_limits('main-domain11.tld', -5)
            1
            >>> mails_db.fetch_domain('main-domain11.tld')['out_limit']
            -1
            >>> mails_db.set_domain_limits('Main-Domain11.tld', 222)
            1
            >>> mails_db.fetch_domain('Main-Domain11.tld')['out_limit']
            222
        """
        out_ph, out_limit = self._normalize_limit(out_limit, 'out_limit')
        return self.run("UPDATE domains SET out_limit = %s WHERE lower(name) = lower(:name)" % out_ph,
                        out_limit=out_limit, name=domain_name).rowcount

    def set_mailname_limits(self, mail_name, domain_name, out_limit=None):
        """ Set in and out limits for mailname. 
            Limit values will be cast to integer. If limit value is None, it will be left untouched.
            Returns number of affected rows.

            >>> mails_db.create_subscription('main-domain12.tld')
            2
            >>> mails_db.create_mailname('mail', 'main-domain12.tld')
            1
            >>> mails_db.set_mailname_limits('mail', 'main-domain12.tld', '0')
            1
            >>> mails_db.fetch_mailname('mail', 'main-domain12.tld')['out_limit']
            0
            >>> mails_db.set_mailname_limits('mail', 'main-domain12.tld', 123)
            1
            >>> mails_db.fetch_mailname('mail', 'main-domain12.tld')['out_limit']
            123
            >>> mails_db.set_mailname_limits('Mail', 'Main-Domain12.tld', 456)
            1
            >>> mails_db.fetch_mailname('Mail', 'Main-Domain12.tld')['out_limit']
            456
        """
        out_ph, out_limit = self._normalize_limit(out_limit, 'out_limit')
        return self.run("UPDATE mails SET out_limit = %s "
                        "WHERE lower(name) = lower(:mail_name) AND domain_id = "
                        "  (SELECT id FROM domains WHERE lower(name) = lower(:domain_name))" % out_ph,
                        out_limit=out_limit, mail_name=mail_name, domain_name=domain_name).rowcount

    def add_user_to_whitelist(self, user_name, uid=None, skip_missing=False):
        """ Add given system user to whitelist.
            Returns number of inserted rows.

            >>> mails_db.add_user_to_whitelist('sysuser13', 213)
            1
            >>> mails_db.is_user_in_whitelist('sysuser13')
            True
            >>> mails_db.is_user_in_whitelist(213)
            True
        """
        if uid is None:
            try:
                uid = pwd.getpwnam(user_name).pw_uid
            except KeyError:
                if skip_missing:
                    log.debug("User `%s' doesn't exist, skip adding to whitelist.", user_name)
                    return 0
                else:
                    raise
        return self.run("INSERT INTO whitelist (id, name) VALUES (?, ?)", uid, user_name).rowcount

    def remove_user_from_whitelist(self, user_name):
        """ Remove given system user from whitelist.
            Returns number of removed rows.

            >>> mails_db.remove_user_from_whitelist('sysuser14')
            0
            >>> mails_db.add_user_to_whitelist('sysuser14', 214)
            1
            >>> mails_db.remove_user_from_whitelist('sysuser14')
            1
            >>> mails_db.is_user_in_whitelist('sysuser14')
            False
        """
        return self.run("DELETE FROM whitelist WHERE name = ?", user_name).rowcount

    def is_user_in_whitelist_by_uid(self, uid):
        """ Check whether a given system user is in whitelist.
            Returns True in case user is in whitelist, False otherwise.

            >>> mails_db.is_user_in_whitelist_by_uid(-1)
            False
        """
        return self.run("SELECT id FROM whitelist WHERE id = ?", uid).fetchone() is not None

    def is_user_in_whitelist_by_name(self, user_name):
        """ Check whether a given system user is in whitelist.
            Returns True in case user is in whitelist, False otherwise.

            >>> mails_db.is_user_in_whitelist_by_name('non-existent-sysuser')
            False
        """
        return self.run("SELECT id FROM whitelist WHERE name = ?", user_name).fetchone() is not None

    def is_user_in_whitelist(self, user):
        """ Check whether a given system user is in whitelist.
            'user' argument can be either uid (integer) or user name (string).
            Returns True in case user is in whitelist, False otherwise.
        """
        if isinstance(user, int):
            return self.is_user_in_whitelist_by_uid(user)
        else:
            return self.is_user_in_whitelist_by_name(user)

    def get_whitelist_users(self):
        """ Return names of whitelisted users.

            >>> mails_db.add_user_to_whitelist('sysuser15', 215)
            1
            >>> 'sysuser15' in mails_db.get_whitelist_users()
            True
            >>> for x in mails_db.get_whitelist_users():
            ...     _ = mails_db.remove_user_from_whitelist(x)
            >>> mails_db.add_user_to_whitelist('sysuser16', 216)
            1
            >>> mails_db.add_user_to_whitelist('sysuser15', 215)
            1
            >>> mails_db.get_whitelist_users() == ['sysuser15', 'sysuser16']
            True
        """
        return [x['name'] for x in self.run("SELECT name FROM whitelist ORDER BY name").fetchall()]

    def _populate_whitelist_from(self, users):
        """ Populate whitelist with system users who are allowed to send mails by default. """
        for user in users:
            self.add_user_to_whitelist(user, skip_missing=True)

class MailsAliasesDb(MailsDb):
    ALIASES_DB_SCHEMA = strip_re("""
        CREATE TABLE mail_aliases (
            id              INTEGER PRIMARY KEY,
            name            VARCHAR NOT NULL,               -- mail alias name (part before '@')
            mail_id         INTEGER NOT NULL,

            UNIQUE (name, mail_id),
            FOREIGN KEY(mail_id) REFERENCES mails(id)
                ON UPDATE CASCADE
                ON DELETE CASCADE
        );

        CREATE INDEX idx_mail_aliases_fk ON mail_aliases(mail_id);

        CREATE TABLE domain_aliases (
            id              INTEGER PRIMARY KEY,
            name            VARCHAR NOT NULL UNIQUE,        -- domain alias name
            domain_id       INTEGER NOT NULL,

            FOREIGN KEY(domain_id) REFERENCES domains(id)
                ON UPDATE CASCADE
                ON DELETE CASCADE
        );

        CREATE INDEX idx_domain_aliases_fk ON domain_aliases(domain_id);
    """, r'^ {8}')

    def __init__(self, db_path=None):
        super(MailsAliasesDb, self).__init__(db_path)

    def _get_schema(self):
        schema = super(MailsAliasesDb, self)._get_schema()
        schema += self.ALIASES_DB_SCHEMA
        return schema

    def fetch_domain_aliases(self, domain_name):
        raise NotImplementedError("fetch_domain_aliases() is not implemented")

    def fetch_mailname_aliases(self, mail_name, domain_name):
        raise NotImplementedError("fetch_mailname_aliases() is not implemented")

    def _get_domain_name_by_alias(self, domain_alias):
        """ Returns domain real name by domain alias. If such alias doesn't exist returns alias name.
            
            >>> aliases_db._get_domain_name_by_alias(u'non-existent-domain')
            u'non-existent-domain'
            >>> aliases_db._get_domain_name_by_alias(u'alias-main-domain.tld')
            u'a-main-domain.tld'
            >>> aliases_db._get_domain_name_by_alias(u'Non-Existent-Domain')
            u'non-existent-domain'
            >>> aliases_db._get_domain_name_by_alias(u'Alias-Main-Domain.tld')
            u'a-main-domain.tld'
        """
        domain = self.run("SELECT lower(d.name) AS name FROM domains d JOIN domain_aliases da ON d.id = da.domain_id "
                          "WHERE lower(da.name) = lower(?)", domain_alias).fetchone() or {'name': domain_alias.lower()}
        return domain['name']

    def fetch_domain_by_alias(self, domain_alias):
        """ Fetch domain information by its alias or real name. See fetch_domain() method. 

            >>> aliases_db.fetch_domain_by_alias('non-existent-domain') is None
            True
            >>> aliases_db.fetch_domain_by_alias('alias-main-domain.tld')['name']
            u'a-main-domain.tld'
            >>> aliases_db.fetch_domain_by_alias('a-main-domain.tld')['name']
            u'a-main-domain.tld'
            >>> aliases_db.fetch_domain_by_alias('Alias-Main-Domain.tld')['name']
            u'a-main-domain.tld'
            >>> aliases_db.fetch_domain_by_alias('A-Main-Domain.tld')['name']
            u'a-main-domain.tld'
        """
        return self.fetch_domain(self._get_domain_name_by_alias(domain_alias))

    def fetch_mailname_by_aliases(self, mail_alias, domain_alias):
        """ Fetch mailname information by mail and domain aliases or real names. See fetch_mailname() method.

            >>> aliases_db.fetch_mailname_by_aliases('a-mail', 'non-existent-domain') is None
            True
            >>> aliases_db.fetch_mailname_by_aliases('non-existent-mail', 'alias-main-domain.tld') is None
            True
            >>> aliases_db.fetch_mailname_by_aliases('alias-mail', 'alias-main-domain.tld')['mail_name']
            u'a-mail'
            >>> aliases_db.fetch_mailname_by_aliases('alias-mail', 'alias-main-domain.tld')['domain_name']
            u'a-main-domain.tld'
            >>> aliases_db.fetch_mailname_by_aliases('a-mail', 'a-main-domain.tld')['mail_name']
            u'a-mail'
            >>> aliases_db.fetch_mailname_by_aliases('a-mail', 'a-main-domain.tld')['domain_name']
            u'a-main-domain.tld'

            Note that mail_alias is not unique, but domain_alias is.

            >>> aliases_db.fetch_mailname_by_aliases('alias-mail', 'alias-main-domain-2.tld')['domain_name']
            u'a-main-domain-2.tld'
            >>> aliases_db.fetch_mailname_by_aliases('alias-mail', 'alias-main-domain-2.tld')['mail_name']
            u'a-mail-2'
            >>> aliases_db.fetch_mailname_by_aliases('alias-mail', 'a-main-domain-3.tld')['mail_name']
            u'alias-mail'

            >>> aliases_db.fetch_mailname_by_aliases('Alias-Mail', 'Alias-Main-Domain.tld')['mail_name']
            u'a-mail'
            >>> aliases_db.fetch_mailname_by_aliases('Alias-Mail', 'Alias-Main-Domain.tld')['domain_name']
            u'a-main-domain.tld'
        """
        domain_name = self._get_domain_name_by_alias(domain_alias)
        mail = self.run("SELECT lower(m.name) AS name "
                        "FROM mails m JOIN domains d ON m.domain_id = d.id JOIN mail_aliases ma ON m.id = ma.mail_id "
                        "WHERE lower(ma.name) = lower(?) AND d.name = ?",
                        mail_alias, domain_name).fetchone() or {'name': mail_alias}
        return self.fetch_mailname(mail['name'], domain_name)

    def list_mailnames_by_local_alias(self, mail_alias):
        """ Fetch mailname information for all matching addresses by mail local part alias or real name.
            See list_mailnames_by_local_part() method.

            This method assumes that mailname is never aliased to itself (i.e. real and alias names cannot be same).

            >>> aliases_db.list_mailnames_by_local_alias('non-existent-mail')
            []
            >>> map(lambda e: skip_key(e, '_id'), aliases_db.list_mailnames_by_local_alias('alias-mail')) == [
            ...         {'mail_name': 'a-mail',     'domain_name': 'a-main-domain.tld',   'out_limit': -1},
            ...         {'mail_name': 'a-mail-2',   'domain_name': 'a-main-domain-2.tld', 'out_limit': -1},
            ...         {'mail_name': 'alias-mail', 'domain_name': 'a-main-domain-3.tld', 'out_limit': -1},
            ...     ]
            True
            >>> map(lambda e: skip_key(e, '_id'), aliases_db.list_mailnames_by_local_alias('Alias-Mail')) == [
            ...         {'mail_name': 'a-mail',     'domain_name': 'a-main-domain.tld',   'out_limit': -1},
            ...         {'mail_name': 'a-mail-2',   'domain_name': 'a-main-domain-2.tld', 'out_limit': -1},
            ...         {'mail_name': 'alias-mail', 'domain_name': 'a-main-domain-3.tld', 'out_limit': -1},
            ...     ]
            True
        """
        return self.run("SELECT m.id AS _id, m.name AS mail_name, d.name AS domain_name, m.out_limit "
                        "FROM mails m JOIN domains d ON m.domain_id = d.id JOIN mail_aliases ma ON m.id = ma.mail_id "
                        "WHERE lower(ma.name) = lower(?)", mail_alias).fetchall() + self.list_mailnames_by_local_part(mail_alias)

    def create_domain_alias(self, domain_alias, domain_name):
        """ Create new domain alias.
            Returns number of inserted rows.

            >>> aliases_db.create_domain_alias('alias2.tld', 'a-main-domain.tld')
            1
            >>> aliases_db.create_domain_alias('Alias-MixedCase2.tld', 'A-Main-Domain.tld')
            1
        """
        return self.run("INSERT INTO domain_aliases (name, domain_id) "
                        "VALUES (lower(?), (SELECT id FROM domains WHERE lower(name) = lower(?)))", domain_alias, domain_name).rowcount

    def create_mailname_alias(self, mail_alias, mail_name, domain_name):
        """ Create new mailname alias.
            Returns number of inserted rows.

            >>> aliases_db.create_mailname_alias('alias2-mail', 'a-mail', 'a-main-domain.tld')
            1
            >>> aliases_db.create_mailname_alias('Alias-MixedCase2-Mail', 'A-Mail', 'A-Main-Domain.tld')
            1
        """
        return self.run("INSERT INTO mail_aliases (name, mail_id) "
                        "VALUES (lower(:mail_alias), (SELECT m.id FROM mails m JOIN domains d ON m.domain_id = d.id "
                                              "WHERE lower(m.name) = lower(:mail_name) AND lower(d.name) = lower(:domain_name)))",
                        mail_alias=mail_alias, mail_name=mail_name, domain_name=domain_name).rowcount

    def delete_domain_alias(self, domain_alias):
        """ Delete domain alias.
            Returns number of removed rows.

            >>> aliases_db.create_domain_alias('alias3.tld', 'a-main-domain.tld')
            1
            >>> aliases_db.delete_domain_alias('alias3.tld')
            1
            >>> aliases_db.fetch_domain_by_alias('alias3.tld') is None
            True

            >>> aliases_db.create_domain_alias('Alias-MixedCase3.tld', 'A-Main-Domain.tld')
            1
            >>> aliases_db.delete_domain_alias('Alias-MixedCase3.tld')
            1
            >>> aliases_db.fetch_domain_by_alias('Alias-MixedCase3.tld') is None
            True
        """
        return self.run("DELETE FROM domain_aliases WHERE lower(name) = lower(?)", domain_alias).rowcount

    def delete_mailname_alias(self, mail_alias, domain_name):
        """ Delete mailname alias.
            Returns number of removed rows.

            >>> aliases_db.create_mailname_alias('alias4-mail', 'a-mail', 'a-main-domain.tld')
            1
            >>> aliases_db.delete_mailname_alias('alias4-mail', 'a-main-domain.tld')
            1
            >>> aliases_db.fetch_mailname_by_aliases('alias4-mail', 'a-main-domain.tld') is None
            True

            >>> aliases_db.create_mailname_alias('Alias-MixedCase4-Mail', 'A-Mail', 'A-Main-Domain.tld')
            1
            >>> aliases_db.delete_mailname_alias('Alias-MixedCase4-Mail', 'A-Main-Domain.tld')
            1
            >>> aliases_db.fetch_mailname_by_aliases('Alias-MixedCase4-Mail', 'A-Main-Domain.tld') is None
            True
        """
        return self.run("DELETE FROM mail_aliases WHERE lower(name) = lower(:mail_alias) AND mail_id IN "
                        "  (SELECT m.id FROM mails m JOIN domains d ON m.domain_id = d.id "
                        "   WHERE lower(d.name) = lower(:domain_name))",
                        mail_alias=mail_alias, domain_name=domain_name).rowcount

    def rename_domain_alias(self, old_domain_alias, new_domain_alias):
        """ Rename existing domain alias.
            Returns number of affected rows.

            >>> aliases_db.create_domain_alias('alias5.tld', 'a-main-domain.tld')
            1
            >>> aliases_db.rename_domain_alias('alias5.tld', 'renamed-alias5.tld')
            1
            >>> aliases_db.fetch_domain_by_alias('alias5.tld') is None
            True
            >>> aliases_db.fetch_domain_by_alias('renamed-alias5.tld')['name']
            u'a-main-domain.tld'

            >>> aliases_db.create_domain_alias('Alias-MixedCase5.tld', 'A-Main-Domain.tld')
            1
            >>> aliases_db.rename_domain_alias('Alias-MixedCase5.tld', 'Renamed-Alias-MixedCase5.tld')
            1
            >>> aliases_db.fetch_domain_by_alias('Alias-MixedCase5.tld') is None
            True
            >>> aliases_db.fetch_domain_by_alias('Renamed-Alias-MixedCase5.tld')['name']
            u'a-main-domain.tld'
        """
        return self.run("UPDATE domain_aliases SET name = lower(?) WHERE lower(name) = lower(?)",
                        new_domain_alias, old_domain_alias).rowcount

    def rename_mailname_alias(self, old_mail_alias, new_mail_alias, domain_name):
        raise NotImplementedError("rename_mailname_alias() is not implemented")


class MailsAliasesVhostidDb(MailsAliasesDb):
    VHOST_IDS_DB_SCHEMA = strip_re("""
        -- vhost-id is always a GUID, but we don't really care how it actually looks
        CREATE TABLE vhost_ids (
            guid            VARCHAR(36) NOT NULL PRIMARY KEY,
            domain_id       INTEGER NOT NULL,

            FOREIGN KEY(domain_id) REFERENCES domains(id)
                ON UPDATE CASCADE
                ON DELETE CASCADE
        );

        CREATE INDEX idx_vhost_ids_fk ON vhost_ids(domain_id);
    """, r'^ {8}')

    def __init__(self, db_path=None):
        super(MailsAliasesVhostidDb, self).__init__(db_path)

    def _get_schema(self):
        schema = super(MailsAliasesVhostidDb, self)._get_schema()
        schema += self.VHOST_IDS_DB_SCHEMA
        return schema

    def add_vhost_id(self, guid, domain_name):
        """ Add new vhost-id for a given domain.
            Returns number of inserted rows.

            >>> vhostid_db.add_vhost_id('6288760a-b0c5-4ad1-ba1e-faec735bd019', 'v-main-domain.tld')
            1
        """
        return self.run("INSERT INTO vhost_ids (guid, domain_id) "
                        "VALUES (?, (SELECT id FROM domains WHERE lower(name) = lower(?)))",
                        guid, domain_name).rowcount

    def remove_vhost_id(self, guid):
        """ Remove existing vhost-id.
            Returns number of removed rows.

            >>> guid = '00000000-0000-0000-0000-000000000000'
            >>> vhostid_db.add_vhost_id(guid, 'v-main-domain.tld')
            1
            >>> vhostid_db.remove_vhost_id(guid)
            1
            >>> vhostid_db.fetch_domain_by_vhost_id(guid) is None
            True
        """
        return self.run("DELETE FROM vhost_ids WHERE guid = ?", guid).rowcount

    def fetch_domain_by_vhost_id(self, guid):
        """ Fetch domain information by its vhost-id. See fetch_domain() method.

            >>> guid = '00000000-0000-0000-0000-000000000001'
            >>> vhostid_db.add_vhost_id(guid, 'v-main-domain.tld')
            1
            >>> vhostid_db.fetch_domain_by_vhost_id(guid)['name']
            u'v-main-domain.tld'
        """
        domain = self.run("SELECT lower(d.name) AS name FROM domains d JOIN vhost_ids v ON d.id = v.domain_id "
                          "WHERE v.guid = ?", guid).fetchone()
        if domain:
            return self.fetch_domain(domain['name'])
        else:
            return None


class DuplicateRecordError(Exception):
    pass

class MailMessagesDb(MailsAliasesVhostidDb):
    VALID_RESOLUTION_VALUES = ('passed', 'rejected')
    MESSAGES_DB_SCHEMA = strip_re("""
        CREATE TABLE messages (
            message_id      VARCHAR NOT NULL PRIMARY KEY,                   -- X-PPP-Message-ID mail header value
            timestamp       DATETIME NOT NULL DEFAULT CURRENT_TIMESTAMP,    -- last processed at (seconds precision)
            resolution      VARCHAR NOT NULL,                               -- enum {passed, rejected}

            -- references to limits, set if message should be counted towards a given limit
            mail_id         INTEGER DEFAULT NULL,
            domain_id       INTEGER DEFAULT NULL,
            subscription_id INTEGER DEFAULT NULL,
            whitelist_id    INTEGER DEFAULT NULL,

            FOREIGN KEY(mail_id) REFERENCES mails(id)
                ON UPDATE CASCADE
                ON DELETE SET NULL,
            FOREIGN KEY(domain_id) REFERENCES domains(id)
                ON UPDATE CASCADE
                ON DELETE SET NULL,
            FOREIGN KEY(subscription_id) REFERENCES subscriptions(id)
                ON UPDATE CASCADE
                ON DELETE SET NULL,
            FOREIGN KEY(whitelist_id) REFERENCES whitelist(id)
                ON UPDATE CASCADE
                ON DELETE SET NULL
        );

        CREATE INDEX idx_messages_timestamp     ON messages(timestamp);
        CREATE INDEX idx_messages_mails         ON messages(resolution, mail_id,         timestamp);
        CREATE INDEX idx_messages_domains       ON messages(resolution, domain_id,       timestamp);
        CREATE INDEX idx_messages_subscriptions ON messages(resolution, subscription_id, timestamp);
        CREATE INDEX idx_messages_whitelist     ON messages(resolution, whitelist_id,    timestamp);
    """, r'^ {8}')

    NUMBER_OF_UPDATES_PER_CLEANUP = 10000
    NUMBER_OF_CLEANUPS_PER_VACUUM = 24 * 7   # vacuum DB roughly each week

    def __init__(self, db_path=None):
        super(MailMessagesDb, self).__init__(db_path)

    def _get_schema(self):
        schema = super(MailMessagesDb, self)._get_schema()
        schema += self.MESSAGES_DB_SCHEMA
        return schema

    def is_timestamp_valid(self, timestamp):
        """ Check whether a given timestamp string is a valid SQLite timestamp.
            Allowed formats are listed at http://www.sqlite.org/lang_datefunc.html

            >>> mess_db.is_timestamp_valid('now')
            True
            >>> mess_db.is_timestamp_valid('')
            False
            >>> mess_db.is_timestamp_valid(None)
            False
            >>> mess_db.is_timestamp_valid('2013-12-02 15:10:30')
            True
            >>> mess_db.is_timestamp_valid('2013-12-02T15:10:30')
            True
            >>> mess_db.is_timestamp_valid('2013-12-02T21:13:05+07:00')
            True
            >>> mess_db.is_timestamp_valid('Mon, 02 Dec 13 21:17:07 +0700')
            False
            >>> mess_db.is_timestamp_valid('2013-12-02 15:10')
            True
            >>> mess_db.is_timestamp_valid('2013-12-02')
            True
            >>> mess_db.is_timestamp_valid('15:10:30')
            True
        """
        return self.run("SELECT datetime(?) AS ts", timestamp).fetchone()['ts'] is not None

    def _min_timestamp_offset(self, timestamp, offset):
        """ Calculate minimum timestamp between a given one and offset from now.

            >>> from datetime import datetime, timedelta
            >>> fmt = '%Y-%m-%d %H:%M:%S'
            >>> mess_db._min_timestamp_offset('invalid timestamp', '-1 hour')
            Traceback (most recent call last):
                ...
            ValueError: supplied timestamp is invalid: 'invalid timestamp'
            >>> ts = mess_db._min_timestamp_offset('2000-01-02', '-1 hour')
            >>> datetime.strptime(ts, fmt)
            datetime.datetime(2000, 1, 2, 0, 0)
            >>> ts = mess_db._min_timestamp_offset(datetime.utcnow().isoformat(), '-1 hour')
            >>> datetime.utcnow() - datetime.strptime(ts, fmt) >= timedelta(hours=1, seconds=-5)
            True
        """
        if timestamp is None:
            ts = self.run("SELECT datetime('now', ?) AS ts", offset).fetchone()['ts']
        elif not self.is_timestamp_valid(timestamp):
            raise ValueError("supplied timestamp is invalid: '%s'" % timestamp)
        else:
            ts = self.run("SELECT min(datetime(?), datetime('now', ?)) AS ts", timestamp, offset).fetchone()['ts']
        if not ts:
            raise ValueError("either timestamp ('%s') or offset ('%s') is invalid" % (timestamp, offset))
        return ts

    def cleanup_messages(self, up_to_timestamp='now', force=False, allow_vacuuming=False, offset=None):
        """ Remove all entries from messages table with timestamp up to (but not including) up_to_timestamp.
            If not forced, at least 1 hour of records will be always retained. 
            Returns number of removed records.

            >>> def populate(db):
            ...     with db.get_connection() as conn:
            ...         conn.executemany("INSERT INTO messages (message_id, resolution, timestamp) VALUES "
            ...                                               "(?, 'rejected', datetime('now', ?))",
            ...                          [('dummy1', '-1 day'), ('dummy2', '-2 hour'), ('dummy3', '-1 second')] )
            >>> def count(db):
            ...     return db.run("SELECT count(*) AS count FROM messages").fetchone()['count']
            >>> populate(mess_db)
            >>> count(mess_db)
            3
            >>> mess_db.cleanup_messages('now')
            2
            >>> count(mess_db)
            1
            >>> mess_db.cleanup_messages('now', force=True)
            1
            >>> count(mess_db)
            0

            Alternatively, a negative offset from 'now' may be specified and all records up to it will be removed. 

            >>> populate(mess_db)
            >>> mess_db.cleanup_messages(offset='-12 hours')
            1
            >>> count(mess_db)
            2

            Passing allow_vacuuming == True will permit performing periodic SQLite DB compactization, passing
            force == True will force it as well.

            >>> mess_db.cleanup_messages(offset='-12 hours', force=False, allow_vacuuming=True)
            0
            >>> mess_db.cleanup_messages(force=True, allow_vacuuming=True)
            2
        """
        if not offset:
            offset = '-1 hour' if not force else '0 hour'
        up_to = self._min_timestamp_offset(up_to_timestamp, offset)

        log.debug("Removing all messages records up to %s", up_to)
        # WARNING: don't change this query without checking performance via _test_mail_messages_db_performance()
        changed_rows = self.run("DELETE FROM messages WHERE timestamp < datetime(?)", up_to).rowcount

        if allow_vacuuming and (force or random.randrange(self.NUMBER_OF_CLEANUPS_PER_VACUUM) == 0):
            self.vacuum()

        return changed_rows

    def update_message(self, message_id, resolution, 
                       mail_id=None, domain_id=None, subscription_id=None, whitelist_id=None, unique=False):
        """ Insert or update existing record into messages table.
            Record is identified with message_id (usually "X-PPP-Message-ID:" mail header value).
            Other arguments specify resolution ("passed" or "rejected") regarding this message and 
            ids of objects that affected this resolution. Timestamp is always updated to a current one.
            Sometimes this method will also perform messages table maintenance by removing old records.
            Returns number of affected rows (including ones removed during maintenance).

            >>> mess_db.update_message('msgid0', 'junk')
            Traceback (most recent call last):
                ...
            ValueError: resolution value 'junk' is invalid: expected one of ('passed', 'rejected')
            >>> mess_db.update_message('msgid', 'rejected')
            1
            >>> mess_db.update_message('msgid', 'passed')
            1
            >>> mess_db.update_message('msgid', 'rejected', unique=True) # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
                ...
            DuplicateRecordError: column message_id is not unique
            >>> mess_db.run("DELETE FROM messages").rowcount    # beware: this will be 0 on SQLite before 3.6.5
            1
        """
        if resolution not in self.VALID_RESOLUTION_VALUES:
            raise ValueError("resolution value '%s' is invalid: expected one of %s" % 
                             (resolution, self.VALID_RESOLUTION_VALUES))

        params = {
            'message_id': message_id,
            'resolution': resolution,
            'mail_id': mail_id,
            'domain_id': domain_id,
            'subscription_id': subscription_id,
            'whitelist_id': whitelist_id,
        }

        if unique:
            try:
                cursor = self.run(
                    "INSERT INTO messages "
                    "       ( message_id, resolution, mail_id, domain_id, subscription_id, whitelist_id) "
                    "VALUES (:message_id,:resolution,:mail_id,:domain_id,:subscription_id,:whitelist_id)",
                    **params)
            except sqlite3.IntegrityError as ex:
                raise DuplicateRecordError(ex)
        else:
            # Note that we also explicitly specify timestamp - this is to update it 
            # in case message record already exists
            cursor = self.run(
                "INSERT OR REPLACE INTO messages "
                "       ( message_id, resolution, mail_id, domain_id, subscription_id, whitelist_id, timestamp) "
                "VALUES (:message_id,:resolution,:mail_id,:domain_id,:subscription_id,:whitelist_id,datetime('now'))",
                **params)

        changed_rows = cursor.rowcount

        # Self-cleanup to keep DB from bloating even in the absence of scheduled cleanup task
        if cursor.lastrowid % self.NUMBER_OF_UPDATES_PER_CLEANUP == 0:
            # Default statistics collection (BL) is called once an hour and it usually requests data only for 
            # previous full hour. So it is quite safe to remove anything older than 2 hours.
            changed_rows += self.cleanup_messages('now', offset='-2 hour', allow_vacuuming=False)

        return changed_rows

    def count_messages(self, object_type, object_id, resolution='passed'):
        """ Count number of messages with a given resolution that are no more than 1 hour old, that are 
            associated with a given object defined by object_type and its id (object_id).

            >>> db = MailMessagesDb(Sqlite3Db.IN_MEMORY)
            >>> _create_db_entities(db, 1)
            >>> _ = db.update_message('msgid0', resolution='passed', mail_id=1, domain_id=1, subscription_id=1)
            >>> _ = db.update_message('msgid1', resolution='passed',            domain_id=1, subscription_id=1)
            >>> _ = db.update_message('msgid2', resolution='passed',                         subscription_id=1)
            >>> _ = db.update_message('msgid3', resolution='rejected', mail_id=1)
            >>> _ = db.update_message('msgid4', resolution='rejected', domain_id=1)
            >>> _ = db.update_message('msgid5', resolution='rejected', subscription_id=1)
            >>> _ = db.update_message('msgid6', resolution='passed')
            >>> _ = db.update_message('msgid7', resolution='passed', whitelist_id=1)
            >>> types = ('mail', 'domain', 'subscription', 'whitelist')
            >>> [db.count_messages(object_type, 1) for object_type in types]
            [1, 2, 3, 1]
            >>> [db.count_messages(object_type, 1, 'rejected') for object_type in types]
            [1, 1, 1, 0]
        """
        field_names = {
            'whitelist': 'whitelist_id',
            'mail': 'mail_id',
            'domain': 'domain_id',
            'subscription': 'subscription_id',
        }

        if resolution not in self.VALID_RESOLUTION_VALUES:
            raise ValueError("resolution value '%s' is invalid: expected one of %s" % 
                             (resolution, self.VALID_RESOLUTION_VALUES))

        if object_type not in field_names.keys():
            raise ValueError("object type value '%s' is invalid: expected one of %s" % 
                             (object_type, field_names.keys()))

        # WARNING: don't change this query without checking performance via _test_mail_messages_db_performance()
        return self.run("SELECT count(*) AS count FROM messages "
                        "WHERE resolution = ? "
                        "   AND %s = ? "
                        "   AND timestamp > datetime('now', '-1 hour')" % field_names[object_type], 
                        resolution, object_id).fetchone()['count']


class OutgoingMailLimitsDb(MailMessagesDb):
    def __init__(self, db_path=None):
        super(OutgoingMailLimitsDb, self).__init__(db_path)

    def _get_path(self):
        return layout.OUTGOING_MAIL_DB

    def _get_permissions(self):
        popuser = pwd.getpwnam(layout.POPUSER)
        return 0, popuser.pw_gid, 0660

    def _post_create(self):
        super(OutgoingMailLimitsDb, self)._post_create()
        settings = OutgoingMailSettings()
        self._populate_whitelist_from(settings.get_extra_whitelist())


def _create_db_entities(db, amount):
    """ Create amount entities in db: subscriptions, domains, mails, whitelisted users.
        Total number of entities will be about amount^3.
    """
    for i in xrange(amount):
        main_domain_name = 'main-domain-%03d.tld' % i
        db.create_subscription(main_domain_name)

        sysuser_name = 'sysuser-%03d' % i
        db.create_sysuser(sysuser_name, main_domain_name, i)

        for j in xrange(amount):
            if j != 0:
                domain_name = 'domain-%03d-%03d.tld' % (i, j)
                db.create_domain(domain_name, main_domain_name)
            else:
                domain_name = main_domain_name

            for k in xrange(amount):
                mail_name = 'mail-%03d-%03d-%03d' % (i, j, k)
                db.create_mailname(mail_name, domain_name)

    for i in xrange(amount):
        db.add_user_to_whitelist('system-user-%03d' % i, i + 1)

def _test_mail_messages_db_performance(amount=5, messages=20000, executions=5000):
    def map_to_ids(data):
        return [None] + map(lambda d: d['id'], data)

    print("Running count_messages() and cleanup_messages() performance test. This will take a while.")
    db = MailMessagesDb('.test-outgoing-mail-perf.db')
    
    try:
        # Populate DB
        print(" Populating DB with objects (%d)..." % amount)
        _create_db_entities(db, amount)

        subscriptions = map_to_ids(db.run("SELECT id FROM subscriptions").fetchall())
        domains       = map_to_ids(db.run("SELECT id FROM domains").fetchall())
        mails         = map_to_ids(db.run("SELECT id FROM mails").fetchall())
        whitelist     = map_to_ids(db.run("SELECT id FROM whitelist").fetchall())

        print(" Populating DB with messages (%d)..." % messages)
        random.seed(100500)
        with db.get_connection() as conn:
            conn.executemany(
                "INSERT OR REPLACE INTO messages "
                "       ( message_id, resolution, mail_id, domain_id, subscription_id, whitelist_id, timestamp) "
                "VALUES (:message_id,:resolution,:mail_id,:domain_id,:subscription_id,:whitelist_id, "
                                                                                        " datetime('now', :offset))",
                ({
                    'message_id':       'message-id-%06d' % i,
                    'resolution':       random.choice(MailMessagesDb.VALID_RESOLUTION_VALUES),
                    'mail_id':          random.choice(mails),
                    'domain_id':        random.choice(domains),
                    'subscription_id':  random.choice(subscriptions),
                    'whitelist_id':     random.choice(whitelist),
                    'offset':           '%f hour' % random.uniform(-2.0, 0.0),
                } for i in xrange(messages))
            )

        # Time count_messages() and cleanup_messages() calls
        objects = {
            'subscription': subscriptions[1:],
            'domain': domains[1:],
            'mail': mails[1:],
            'whitelist': whitelist[1:],
        }

        from timeit import timeit, repeat
        result = {}
        for object_type, ids in objects.items():
            print(" Timing count_messages() for %s objects (total %d in DB)..." % (object_type, len(ids)))
            result[object_type] = repeat(lambda: db.count_messages(object_type, random.choice(ids)),
                                         number=executions)
        print(" Timing cleanup_messages() among %d records..." % messages)
        result['cleanup'] = [timeit(lambda: db.cleanup_messages(), number=1)]

        print(" Performance values (3 * %d executions for each object): %s" % (executions, result))
        return sum(map(min, result.values()))
    finally:
        # Remove DB file after test
        db.remove()

if __name__ == '__main__':
    import doctest

    def skip_key(dictionary, key):
        dictionary.pop(key, None)
        return dictionary

    mails_db = MailsDb(Sqlite3Db.IN_MEMORY)

    aliases_db = MailsAliasesDb(Sqlite3Db.IN_MEMORY)
    aliases_db.create_subscription('a-main-domain.tld')
    aliases_db.create_sysuser('a-sysuser', 'a-main-domain.tld', 100)
    aliases_db.create_mailname('a-mail', 'a-main-domain.tld')
    aliases_db.create_mailname_alias('alias-mail', 'a-mail', 'a-main-domain.tld')
    aliases_db.create_domain_alias('alias-main-domain.tld', 'a-main-domain.tld')
    aliases_db.create_subscription('a-main-domain-2.tld')
    aliases_db.create_mailname('a-mail-2', 'a-main-domain-2.tld')
    aliases_db.create_mailname_alias('alias-mail', 'a-mail-2', 'a-main-domain-2.tld')
    aliases_db.create_domain_alias('alias-main-domain-2.tld', 'a-main-domain-2.tld')
    aliases_db.create_subscription('a-main-domain-3.tld')
    aliases_db.create_mailname('alias-mail', 'a-main-domain-3.tld')

    vhostid_db = MailsAliasesVhostidDb(Sqlite3Db.IN_MEMORY)
    vhostid_db.create_subscription('v-main-domain.tld')

    mess_db = MailMessagesDb(Sqlite3Db.IN_MEMORY)
    #print( "MailMessagesDb performance index (lower is better): %s" % _test_mail_messages_db_performance() )

    out_db = OutgoingMailLimitsDb(Sqlite3Db.IN_MEMORY)

    sys.exit(doctest.testmod().failed)

# vim: ts=4 sts=4 sw=4 et :
