# Copyright 1999-2017. Plesk International GmbH. All rights reserved.
""" Helpers for statistics collection and object enumeration from outgoing mails rate limiting DB. """
import json
from itertools import imap, izip
from collections import defaultdict

from plesk_outgoing_mail_db import MailsDb, MailMessagesDb


class MailsDbEntitiesEnumerator(object):
    """ Class for convenient enumeration of entities from MailsDb. """
    def __init__(self, db):
        self._db = db
        self._cur_subscription = None
        self._cur_domain = None
        self._cur_mailname = None

        if not isinstance(self._db, MailsDb):
            raise ValueError("MailsDbEntitiesEnumerator works only with instances of MailsDb")

    def subscriptions(self):
        """ Generator of all subscriptions in DB. 
            Each element returned is roughly similar to the one returned by MailsDb.fetch_subscription().
        """
        arr = self._db.run("SELECT s.id AS _id, d.name AS main_domain_name, "
                                  "s.out_limit, s.allow_sendmail "
                           "FROM subscriptions s JOIN domains d ON s.main_domain_id = d.id").fetchall()
        for subscription in arr:
            self._cur_subscription = subscription
            yield subscription
        self._cur_subscription = None

    def domains(self, subscription=None):
        """ Generator of all domains in DB belonging to a given or current subscription. 
            Each element returned is roughly similar to the one returned by MailsDb.fetch_domain().
        """
        if subscription is None:
            subscription = self._cur_subscription

        arr = self._db.run("SELECT d.id AS _id, d.name AS name, d.out_limit "
                           "FROM domains d WHERE d.subscription_id = ?", subscription['_id']).fetchall()
        for domain in arr:
            self._cur_domain = domain
            yield domain
        self._cur_domain = None

    def mailnames(self, domain=None):
        """ Generator of all mailnames in DB belonging to a given or current domain. 
            Each element returned is roughly similar to the one returned by MailsDb.fetch_mailname().
        """
        if domain is None:
            domain = self._cur_domain

        arr = self._db.run("SELECT m.id AS _id, m.name AS mail_name, m.out_limit "
                           "FROM mails m WHERE m.domain_id = ?", domain['_id']).fetchall()
        for mailname in arr:
            self._cur_mailname = mailname
            yield mailname
        self._cur_mailname = None

    def get_messages_query_parts_to_filter_by_subscription(self, subscription=None):
        """ Build SQLite query parts to match messages related to a given or current subscription. """
        if subscription is None:
            subscription = self._cur_subscription
        return ("subscription_id = :subscription_id", {'subscription_id': subscription['_id']})

    def get_messages_query_parts_to_filter_by_domain(self, domain=None):
        """ Build SQLite query parts to match messages related to a given or current domain. """
        if domain is None:
            domain = self._cur_domain
        return ("domain_id = :domain_id", {'domain_id': domain['_id']})

    def get_messages_query_parts_to_filter_by_mailname(self, mailname=None):
        """ Build SQLite query parts to match messages related to a given or current mailname. """
        if mailname is None:
            mailname = self._cur_mailname
        return ("mail_id = :mail_id", {'mail_id': mailname['_id']})

class OutgoingMailStatsHelper(object):
    """ Class to collect statistics based on messages table of MailMessagesDb instance.

        To get statistics data just call fetch() method:

        >>> stats = OutgoingMailStatsHelper(db).fetch()
        >>> #pprint(stats)

        Returned data structure will contain information about such objects as subscriptions, domains, and mailnames,
        as well as their limit values. Dictionaries representing these objects are naturally nested.

        >>> fields(stats['subscriptions'][0], 'out_limit', 'allow_sendmail')
        [1000, True]
        >>> fields(stats['subscriptions'][0]['domains'][0], 'name', 'is_main', 'out_limit')
        [u'main-domain.tld', True, 100]
        >>> fields(stats['subscriptions'][0]['domains'][1], 'name', 'is_main', 'out_limit')
        [u'domain.tld', False, -1]
        >>> fields(stats['subscriptions'][0]['domains'][1], 'stats', 'mails')
        [[], []]
        >>> fields(stats['subscriptions'][0]['domains'][0]['mails'][0], 'name', 'out_limit')
        [u'mailname', 10]

        Statistics data is contained within 'stats' field of a given dictionary that represents an object:

        >>> stats['subscriptions'][0]['domains'][0]['mails'][0]['stats'] == [
        ...         {'time': '2000-01-01 12:00', 'passed': 2, 'rejected': 0},
        ...     ]
        True
        >>> stats['subscriptions'][0]['domains'][0]['stats'] == [
        ...         {'time': '2000-01-01 12:00', 'passed': 2, 'rejected': 1},
        ...     ]
        True
        >>> stats['subscriptions'][0]['stats'] == [
        ...         {'time': '2000-01-01 12:00', 'passed': 2, 'rejected': 0},
        ...         {'time': '2000-01-01 12:01', 'passed': 0, 'rejected': 1},
        ...     ]
        True

        Lists with statistics entries are sorted according to 'time' field. However entries for a given
        time segment may be skipped if no records exist for it (i.e. all values should be considered 0).

        >>> subscr_stats_timestamps = map(lambda entry: entry['time'], stats['subscriptions'][0]['stats'])
        >>> sorted(subscr_stats_timestamps) == subscr_stats_timestamps
        True

        Portion of data to fetch is customizable via other class methods, e.g. starting_from():

        >>> stats = OutgoingMailStatsHelper(db).starting_from('12:01:00').skipping_zero_counters().fetch()
        >>> #pprint(stats)
        >>> stats['subscriptions'][0]['stats'] == [{'time': '2000-01-01 12:01', 'rejected': 1}]
        True
        >>> stats['subscriptions'][0]['domains'][0]['stats']
        []
        >>> stats['subscriptions'][0]['domains'][0]['mails'][0]['stats']
        []
    """
    def __init__(self, db):
        self._db = db
        self._from_ts = None
        self._skip_zero_counters = False
        self._skip_limits = False
        self._time_step = 'minute'

    def starting_from(self, timestamp):
        """ Set starting timestamp. All statistics entries will have timestamp no less than this value. 
            Setting to None discards previous changes (default is to include all entries).
        """
        if timestamp is not None and not self._db.is_timestamp_valid(timestamp):
            raise ValueError("'from' timestamp is invalid: '%s'" % timestamp)
        self._from_ts = timestamp
        return self

    def skipping_zero_counters(self, skip=True):
        """ Skip zero counters (such as 'passed' or 'rejected') in statistics entries.
            Don't skip them if False is passed. This is the default behavior.

            >>> stats = OutgoingMailStatsHelper(db).skipping_zero_counters().fetch()
            >>> sorted(stats['subscriptions'][0]['domains'][0]['mails'][0]['stats'][0].keys()) == ['passed', 'time']
            True
        """
        self._skip_zero_counters = skip
        return self

    def skipping_limits(self, skip=True):
        """ Skip limit values (out_limit, allow_sendmail) in output.
            Include them if False is passed. This is the default behavior.

            >>> stats = OutgoingMailStatsHelper(db).skipping_limits().fetch()
            >>> data = stats['subscriptions'][0]
            >>> 'out_limit' not in data, 'allow_sendmail' not in data
            (True, True)
            >>> data = stats['subscriptions'][0]['domains'][0]
            >>> 'out_limit' not in data
            True
            >>> data = stats['subscriptions'][0]['domains'][0]['mails'][0]
            >>> 'out_limit' not in data
            True
        """
        self._skip_limits = skip
        return self

    def with_time_step(self, step='minute'):
        """ Change statistics time segment length.
            Default value is 'minute'.

            >>> stats = OutgoingMailStatsHelper(db).with_time_step('hour').fetch()
            >>> stats['subscriptions'][0]['stats'] == [
            ...         {'time': '2000-01-01 12', 'passed': 2, 'rejected': 1},
            ...     ]
            True
        """
        if step not in ('minute', 'hour'):
            raise ValueError("time step value is invalid: '%s'" % step)
        self._time_step = step
        return self

    def fetch(self):
        """ Fetch statistics data according to currently set criteria. """
        enumerator = MailsDbEntitiesEnumerator(self._db)
        data = {
            'subscriptions': self._fetch_subscriptions(enumerator),
        }

        return data

    def _fetch_mailnames(self, enumerator):
        mailnames_data = []

        for mailname in enumerator.mailnames():
            mailname_data = {
                'name': mailname['mail_name'],
            }

            if not self._skip_limits:
                mailname_data.update(out_limit=mailname['out_limit'])

            filter_expr, filter_params = enumerator.get_messages_query_parts_to_filter_by_mailname()
            mailname_data['stats'] = self._fetch_statistics(filter_expr, filter_params)

            mailnames_data.append(mailname_data)

        return mailnames_data

    def _fetch_domains(self, enumerator, main_domain_name):
        domains_data = []

        for domain in enumerator.domains():
            domain_data = {
                'name': domain['name'],
                'is_main': (domain['name'] == main_domain_name),
            }

            if not self._skip_limits:
                domain_data.update(out_limit=domain['out_limit'])

            filter_expr, filter_params = enumerator.get_messages_query_parts_to_filter_by_domain()
            domain_data['stats'] = self._fetch_statistics(filter_expr, filter_params)

            domain_data['mails'] = self._fetch_mailnames(enumerator)

            domains_data.append(domain_data)

        return domains_data

    def _fetch_subscriptions(self, enumerator):
        subscriptions_data = []

        for subscription in enumerator.subscriptions():
            subscription_data = {}

            if not self._skip_limits:
                subscription_data.update(out_limit=subscription['out_limit'],
                                         allow_sendmail=bool(subscription['allow_sendmail']))

            filter_expr, filter_params = enumerator.get_messages_query_parts_to_filter_by_subscription()
            subscription_data['stats'] = self._fetch_statistics(filter_expr, filter_params)

            subscription_data['domains'] = self._fetch_domains(enumerator, subscription['main_domain_name'])

            subscriptions_data.append(subscription_data)

        return subscriptions_data

    def _fetch_statistics(self, query_filter_expr, query_filter_params):
        """ Fetch statistics data (as list of dictionaries) based on messages table. Uses additional query filter
            arguments to refine search (e.g., search only for statistics for a given domain).
            Statistics entires are sorted by 'time' field.
        """
        time_segment_select_expr, time_segment_group_expr = self._time_step_to_sql_query_parts(self._time_step)
        query = ("SELECT %(ts_select_expr)s AS timestamp, resolution, count(*) AS count "
                 "FROM messages "
                 "WHERE %(obj_filter_expr)s AND %(ts_filter_expr)s "
                 "GROUP BY %(ts_group_expr)s, resolution" % {
                     'ts_select_expr': time_segment_select_expr,
                     'ts_group_expr': time_segment_group_expr,
                     'obj_filter_expr': query_filter_expr,
                     'ts_filter_expr': "timestamp >= datetime(:starting_ts)" if self._from_ts else "1",
                 })
        data = self._db.run(query, starting_ts=self._from_ts, **query_filter_params).fetchall()

        stats = self._create_stats_defaultdict(self._skip_zero_counters)
        for entry in data:
            stats[entry['timestamp']][entry['resolution']] = entry['count']

        entries_iterable = imap(lambda ts, d: d.update(time=ts) or d, *izip(*stats.iteritems())) if stats else []
        return sorted(entries_iterable, key=lambda entry: entry['time'])

    def _time_step_to_sql_query_parts(self, step='minute'):
        """ Return (time segment select expression, time segment group by expression) pair for use in SQLite query to
            messages table so that messages are grouped by timestamps with 'step' granularity.
            E.g. for each minute with default arguments.
        """
        if step == 'minute':
            time_segment_expr = "strftime('%Y-%m-%d %H:%M', timestamp)"
            return time_segment_expr, time_segment_expr
        elif step == 'hour':
            time_segment_expr = "strftime('%Y-%m-%d %H', timestamp)"
            return time_segment_expr, time_segment_expr
        else:
            raise ValueError("time step value is invalid: '%s'" % step)

    def _create_stats_defaultdict(self, skip_zero_fields=False):
        """ Create intermediate defaultdict instance with appropriate default value. 
            
            >>> h = OutgoingMailStatsHelper(db)
            >>> d = h._create_stats_defaultdict(skip_zero_fields=True)
            >>> d['12:00:00']['passed'] = 1
            >>> d == {'12:00:00': {'passed': 1}}
            True
            >>> d = h._create_stats_defaultdict()
            >>> d['12:00:00']['passed'] = 1
            >>> d == {'12:00:00': {
            ...                     'passed': 1,
            ...                     'rejected': 0,
            ...      }}
            True
        """
        if skip_zero_fields:
            return defaultdict(dict)
        else:
            def constant_copy_factory(value):
                def next_copy():
                    return value.copy()
                return next_copy

            default_stats_entry = {
                'passed': 0,
                'rejected': 0,
            }

            return defaultdict(constant_copy_factory(default_stats_entry))

def format_statistics_as_json(stats, pretty_print=False):
    """ Formats statistics dictionary generated by OutgoingMailStatsHelper.fetch() into JSON. """
    if pretty_print:
        return json.dumps(stats, indent=4)
    else:
        return json.dumps(stats, separators=(',', ':'))


if __name__ == '__main__':
    import doctest
    import sys
    # from pprint import pprint

    def fields(dictionary, *keys):
        return map(lambda key: dictionary[key], keys)

    db = MailMessagesDb(MailMessagesDb.IN_MEMORY)
    db.create_subscription('main-domain.tld')
    db.create_domain('domain.tld', 'main-domain.tld')
    db.create_sysuser('sysuser', 'main-domain.tld', 100)
    db.create_mailname('mailname', 'main-domain.tld')
    db.add_user_to_whitelist('whitelisted-user', 200)

    db.set_subscription_limits('main-domain.tld', 1000, True)
    db.set_domain_limits('main-domain.tld', 100)
    db.set_domain_limits('domain.tld', -1)
    db.set_mailname_limits('mailname', 'main-domain.tld', 10)

    subscription_id = db.fetch_subscription('main-domain.tld')['_id']
    domain_id = db.fetch_domain('main-domain.tld')['_id']
    addon_domain_id = db.fetch_domain('domain.tld')['_id']
    mail_id = db.fetch_mailname('mailname', 'main-domain.tld')['_id']
    whitelist_id = 200
    sysuser_id = 100

    messages = [
        (1, 'passed',   True,   True,   True,   False,  '12:00:01'),
        (2, 'passed',   True,   True,   True,   False,  '12:00:02'),
        (3, 'rejected', False,  True,   False,  False,  '12:00:03'),
        (5, 'rejected', False,  False,  True,   False,  '12:01:00'),
        (6, 'passed',   False,  False,  False,  True,   '13:00:00'),
    ]

    def add_message_entry(db, entry):
        db.run("INSERT OR REPLACE INTO messages "
               "( message_id, resolution, mail_id, domain_id, subscription_id, whitelist_id, timestamp) VALUES "
               "(:message_id,:resolution,:mail_id,:domain_id,:subscription_id,:whitelist_id, datetime(:timestamp))",
               message_id       = entry[0],
               resolution       = entry[1],
               mail_id          = (mail_id if entry[2] else None),
               domain_id        = (domain_id if entry[3] else None),
               subscription_id  = (subscription_id if entry[4] else None),
               whitelist_id     = (whitelist_id if entry[5] else None),
               timestamp        = entry[6])

    for message in messages:
        add_message_entry(db, message)

    sys.exit(doctest.testmod().failed)

# vim: ts=4 sts=4 sw=4 et :
