# Copyright 1999-2017. Plesk International GmbH. All rights reserved.
""" SQLite DB wrapper. """
import os
import sys
import errno
import sqlite3

import plesk_log


log = plesk_log.getLogger(__name__)


def dict_factory(cursor, row):
    dictionary = {}
    for index, column in enumerate(cursor.description):
        dictionary[column[0]] = row[index]
    return dictionary

class Sqlite3Db(object):
    IN_MEMORY = ":memory:"

    def __init__(self, db_path=None, schema=None):
        self._connection = None
        self._db_schema = schema and schema or self._get_schema()
        self._db_path = db_path and db_path or self._get_path()
        self.get_connection()

    def _get_schema(self):
        """ Returns DB schema (SQL creation script). Override in subclasses. 

            Technically the value returned is not required to be SQL schema per se,
            it should be a valid SQL to bring empty DB into initial state.
        """
        return ""

    def _get_path(self):
        """ Returns DB file path. Override in subclasses. """
        return None

    def _get_permissions(self):
        """ Returns (uid, gid, mode) tuple with expected access rights of DB file. Override in subclasses. """
        return -1, -1, 0600

    def _setup_connection(self, connection):
        """ Additional SQLite connection setup before DB creation and get_connection(). Override in subclasses. """
        pass

    def _post_create(self):
        """ Additional actions that should be executed after DB is created. Override in subclasses. """
        pass

    def _check_settings(self):
        if not self._db_path:
            raise ValueError("Path to SQLite3 database is not set")
        if not self._db_schema:
            raise ValueError("SQLite3 database schema is not set")

    def _set_db_permissions(self):
        """ Set file permissions on existing DB. """
        # Parent directory permissions also matter (need directory write permissions to write to DB), ignore for now.
        if not self.is_in_memory():
            uid, gid, mode = self._get_permissions()
            os.chown(self._db_path, uid, gid)
            os.chmod(self._db_path, mode)
            log.debug("Permissions on DB '%s' were set to uid=%d, gid=%d, mode=%s", self._db_path, uid, gid, oct(mode))

    def is_in_memory(self):
        return self._db_path == Sqlite3Db.IN_MEMORY

    def close(self):
        """ Close DB connection. """
        if self._connection is not None:
            self._connection.close()
            self._connection = None

    def remove(self):
        """ Destroy and remove DB. """
        self.close()
        if not self.is_in_memory():
            try:
                os.unlink(self._db_path)
            except OSError as ex:
                log.debug("DB '%s' removal failed: %s", self._db_path, ex)
                if ex.errno != errno.ENOENT:
                    raise

    def recreate(self, overwrite_existing=True):
        """ Create or recreate (if overwrite_existing == True) DB. 
            May open new connection to DB. 
        """
        self._check_settings()

        if self.is_in_memory():
            log.debug("DB storage is in memory, recreating DB ignoring overwrite_existing value")
        elif os.path.exists(self._db_path):
            log.debug("DB '%s' already exists", self._db_path)
            if not overwrite_existing:
                return

        log.debug("Creating new DB '%s'", self._db_path)

        self.remove()

        self._connection = sqlite3.connect(self._db_path)
        self._setup_connection(self._connection)
        self._set_db_permissions()
        with self._connection as conn:
            conn.executescript(self._db_schema)

        self._post_create()

        log.debug("New DB '%s' was initialized", self._db_path)

    def get_connection(self):
        """ Returns connection object (sqlite3.Connection). Creates DB if it doesn't exist. """
        if self._connection is None:
            self.recreate(overwrite_existing=False)
            if self._connection is None:
                self._connection = sqlite3.connect(self._db_path)
                self._setup_connection(self._connection)
            self._connection.row_factory = dict_factory
        return self._connection

    def run(self, query, *args, **kwargs):
        """ Execute a single query and return resulting sqlite3.Cursor.
            Both named and positional arguments are supported, but only one type at a time.

            >>> db.run("INSERT INTO meta (name, value) VALUES (?, ?)", 'custom', 2) #doctest: +ELLIPSIS
            <....Cursor object at ...>
            >>> db.run("SELECT * FROM meta WHERE value > :value AND name = :name", name='custom', value=1).fetchall()
            [{'name': u'custom', 'value': 2}]
        """
        if args and kwargs:
            raise RuntimeError("Cannot process both named and positional parameters for SQLite query")
        params = kwargs and kwargs or args
        log.debug("Executing query '%s' with parameters %s", query, params)
        with self.get_connection() as conn:
            return conn.execute(query, params)

    def is_foreign_keys_support_enabled(self):
        """ Check if foreign keys support is present and enabled. 
            
            >>> isinstance(db.is_foreign_keys_support_enabled(), bool)
            True
        """
        result = self.run("PRAGMA foreign_keys").fetchone()
        return result is not None and bool(result['foreign_keys'])

    def vacuum(self):
        """ Rebuild the entire database to reduce its size and avoid some of fragmentation.
            This may be a rather expensive operation.

            >>> db.vacuum()
        """
        log.debug("Vacuuming (compacting) database")
        self.run("VACUUM")


if __name__ == '__main__':
    import doctest
    schema = """
        PRAGMA foreign_keys = ON;

        CREATE TABLE meta (
            name            VARCHAR(255) NOT NULL DEFAULT 'default',
            value           INTEGER NOT NULL DEFAULT 0
        );

        INSERT INTO meta DEFAULT VALUES;
    """
    db = Sqlite3Db(Sqlite3Db.IN_MEMORY, schema)
    print("SQLite library version: %s" % sqlite3.sqlite_version)
    print("Foreign key support enabled: %s" % db.is_foreign_keys_support_enabled())
    sys.exit(doctest.testmod().failed)

# vim: ts=4 sts=4 sw=4 et :
